# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cProfile
import os
import sys
from argparse import SUPPRESS, ArgumentParser, Namespace, RawTextHelpFormatter
from datetime import date
from importlib import import_module
from pathlib import Path
from pkgutil import iter_modules
from types import ModuleType
from typing import Dict, List, Set

from prezzemolo.avl_tree import AVLTree

from rp2.abstract_accounting_method import AbstractAccountingMethod
from rp2.abstract_country import AbstractCountry
from rp2.abstract_report_generator import AbstractReportGenerator
from rp2.accounting_engine import AccountingEngine
from rp2.computed_data import ComputedData
from rp2.configuration import (
    MAX_DATE,
    MIN_DATE,
    REPORT_GENERATOR_PACKAGE,
    Configuration,
)
from rp2.input_data import InputData
from rp2.localization import set_generation_language
from rp2.logger import LOG_FILE, LOGGER
from rp2.ods_parser import open_ods, parse_ods
from rp2.tax_engine import compute_tax

_VERSION: str = "1.5.1"

_ACCOUNTING_METHOD_PACKAGE = "rp2.plugin.accounting_method"


def rp2_main(country: AbstractCountry) -> None:
    if "RP2_ENABLE_PROFILER" in os.environ:
        cProfile.runctx("_rp2_main_internal(country)", globals(), locals())
    else:
        _rp2_main_internal(country)


def _rp2_main_internal(country: AbstractCountry) -> None:  # pylint: disable=too-many-branches
    args: Namespace
    assets: List[str]
    parser: ArgumentParser

    AbstractCountry.type_check("country", country)

    parser = _setup_argument_parser(country)
    args = parser.parse_args()

    set_generation_language(args.generation_language)

    _setup_paths(parser=parser, configuration_file=args.configuration_file, input_file=args.input_file, output_dir=args.output_dir)

    # On certain platforms the mpdecimal system library is missing (see details at: https://github.com/eprbell/rp2/issues/25).
    try:
        import _decimal as _  # pylint: disable=import-outside-toplevel
    except ImportError:
        LOGGER.error("The `mpdecimal` library is a required system dependency for RP2, but Python was not able to locate it.")
        sys.exit(1)

    try:
        LOGGER.info("Country: %s", country.country_iso_code)
        LOGGER.info("Generation Language: %s", args.generation_language)

        configuration: Configuration = Configuration(
            configuration_path=args.configuration_file,
            country=country,
            from_date=args.from_date,
            to_date=args.to_date,
            allow_negative_balances=args.allow_negative_balances,
        )
        LOGGER.debug("Configuration object: %s", configuration)

        years_2_accounting_method_names: Dict[int, str] = configuration.years_2_accounting_method_names
        if args.method and configuration.years_2_accounting_method_names:
            LOGGER.error(
                "Accounting method cannot be defined both via -m command line option and 'accounting_methods' section in configuration file: "
                "use only one of them."
            )
            sys.exit(1)
        elif not args.method and configuration.years_2_accounting_method_names:
            years_2_accounting_method_names = configuration.years_2_accounting_method_names
        elif args.method and not configuration.years_2_accounting_method_names:
            years_2_accounting_method_names = {MIN_DATE.year: args.method}
        else:  # neither is defined
            years_2_accounting_method_names = {MIN_DATE.year: country.get_default_accounting_method()}

        old_year: int = MIN_DATE.year
        years_2_accounting_methods: AVLTree[int, AbstractAccountingMethod] = AVLTree()
        for year, accounting_method_name in years_2_accounting_method_names.items():
            try:
                accounting_method_module: ModuleType = import_module(
                    f"{_ACCOUNTING_METHOD_PACKAGE}.{accounting_method_name}", package=_ACCOUNTING_METHOD_PACKAGE
                )
            except ModuleNotFoundError:
                LOGGER.error("Invalid/unsupported accounting method: %s", accounting_method_name)
                sys.exit(1)
            if not hasattr(accounting_method_module, "AccountingMethod"):
                LOGGER.error("Accounting method plugin %s doesn't have an AccountingMethod class", accounting_method_name)
                sys.exit(1)
            accounting_method: AbstractAccountingMethod = accounting_method_module.AccountingMethod()
            if len(years_2_accounting_method_names) == 1:
                LOGGER.info("Accounting method: %s", accounting_method_name)
            else:
                if year - old_year > 1:
                    LOGGER.info("Accounting method for %s->%s: %s", old_year, year, accounting_method_name)
                else:
                    LOGGER.info("Accounting method for %s: %s", year, accounting_method_name)
            years_2_accounting_methods.insert_node(year, accounting_method)
            old_year = year

        accounting_engine: AccountingEngine = AccountingEngine(years_2_methods=years_2_accounting_methods)

        LOGGER.info("Configuration file: %s", args.configuration_file)

        if args.plugin:
            LOGGER.error("Command line option '-l' or '--plugin' has been deprecated: use the 'generators' section in the configuration file instead.")
            sys.exit(1)

        if args.asset:
            assets = [args.asset]
        else:
            assets = list(configuration.assets)
        assets.sort()

        asset_to_computed_data: Dict[str, ComputedData] = {}
        asset: str

        LOGGER.info("Input file: %s", args.input_file)
        input_file_handle: object = open_ods(configuration=configuration, input_file_path=args.input_file)
        for asset in assets:
            LOGGER.info("Processing %s", asset)

            input_data: InputData = parse_ods(configuration=configuration, asset=asset, input_file_handle=input_file_handle)
            LOGGER.debug("InputData object: %s", input_data)

            computed_data: ComputedData = compute_tax(configuration=configuration, accounting_engine=accounting_engine, input_data=input_data)
            LOGGER.debug("ComputedData object: %s", computed_data)

            asset_to_computed_data[asset] = computed_data

        # Run report generators (both country-specific and non-country-specific)
        _find_and_run_report_generators(
            configuration=configuration,
            package_paths=[REPORT_GENERATOR_PACKAGE, f"{REPORT_GENERATOR_PACKAGE}.{country.country_iso_code}"],
            args=args,
            country=country,
            years_2_accounting_method_names=years_2_accounting_method_names,
            asset_to_computed_data=asset_to_computed_data,
            from_date=configuration.from_date,
            to_date=configuration.to_date,
        )
    except Exception:  # pylint: disable=broad-except
        LOGGER.exception("Fatal exception occurred:")
        sys.exit(1)

    LOGGER.info("Log file: %s", LOG_FILE)
    LOGGER.info("Generated output directory: %s", args.output_dir)
    LOGGER.info("Done")


def _find_and_run_report_generators(
    configuration: Configuration,
    package_paths: List[str],
    args: Namespace,
    country: AbstractCountry,
    years_2_accounting_method_names: Dict[int, str],
    asset_to_computed_data: Dict[str, ComputedData],
    from_date: date,
    to_date: date,
) -> None:
    generators = configuration.generators.copy()
    for package_path in package_paths:
        # Load report generator plugins and call their generate() method
        try:
            package: ModuleType = import_module(package_path)
        except ModuleNotFoundError:
            # Path not found
            continue
        plugin_name: str
        is_package: bool
        for *_, plugin_name, is_package in iter_modules(package.__path__, package.__name__ + "."):
            if is_package:
                continue
            if plugin_name not in generators:
                continue
            generators.remove(plugin_name)
            output_module: ModuleType = import_module(plugin_name, package=REPORT_GENERATOR_PACKAGE)
            if hasattr(output_module, "Generator"):
                generator: AbstractReportGenerator = output_module.Generator()
                LOGGER.debug("Generator object: '%s'", generator)
                LOGGER.info("Generating output for plugin '%s'", plugin_name)
                if not hasattr(generator, "generate"):
                    LOGGER.error("Plugin '%s' has no 'generate' method. Exiting...", plugin_name)
                    sys.exit(1)
                generator.generate(
                    country=country,
                    years_2_accounting_method_names=years_2_accounting_method_names,
                    asset_to_computed_data=asset_to_computed_data,
                    output_dir_path=args.output_dir,
                    output_file_prefix=args.prefix,
                    from_date=from_date,
                    to_date=to_date,
                    generation_language=args.generation_language,
                )

    if generators:
        LOGGER.error("Report generator plugins %s not found. Exiting...", ", ".join(generators))
        sys.exit(1)


def _validate_accounting_methods(country: AbstractCountry) -> List[str]:
    # Load accounting method plugins
    package: ModuleType = import_module(_ACCOUNTING_METHOD_PACKAGE)
    plugin_name: str
    is_package: bool
    result: List[str] = []
    accounting_methods: Set[str] = country.get_accounting_methods()

    if not country.get_default_accounting_method():
        LOGGER.error("No default accounting method defined in the country plugin. Exiting...")
        sys.exit(1)

    if not country.get_accounting_methods():
        LOGGER.error("No accounting methods defined in the country plugin. Exiting...")
        sys.exit(1)

    for *_, plugin_name, is_package in iter_modules(package.__path__, package.__name__ + "."):
        if is_package:
            continue
        normalized_plugin_name = plugin_name.rsplit(".", 1)[1]
        if normalized_plugin_name in accounting_methods:
            result.append(normalized_plugin_name)

    if not result:
        LOGGER.error("No accounting method plugins found. Exiting...")
        sys.exit(1)

    return sorted(result)


def _setup_argument_parser(country: AbstractCountry) -> ArgumentParser:
    accounting_methods = _validate_accounting_methods(country)

    parser: ArgumentParser = ArgumentParser(
        description=(
            "Generate capital gain/loss report and balances for crypto holdings. Links:\n"
            "- documentation: https://github.com/eprbell/rp2/blob/main/README.md\n"
            "- FAQ: https://github.com/eprbell/rp2/blob/main/docs/user_faq.md\n"
            "- support RP2 by leaving a star on Github: https://github.com/eprbell/rp2"
        ),
        formatter_class=RawTextHelpFormatter,
    )

    parser.add_argument(
        "-a",
        "--asset",
        action="store",
        help="Generate report only for the given ASSET",
        metavar="ASSET",
        type=str,
    )
    parser.add_argument(
        "-f",
        "--from_date",
        action="store",
        default=MIN_DATE,
        help="Generate report from the given date (in ISO 8601 format: e.g. YYYY-MM-DD)",
        metavar="DATE",
        type=date.fromisoformat,
    )
    parser.add_argument(
        "-g",
        "--generation-language",
        action="store",
        default=country.get_default_generation_language(),
        help="Language to use during generation (in ISO 639-1 format)",
        metavar="GENERATION_LANGUAGE",
        type=str,
    )
    parser.add_argument(
        "-l",
        "--plugin",
        action="store",
        help=SUPPRESS,
        metavar="PLUGIN",
        type=str,
    )
    parser.add_argument(
        "-m",
        "--method",
        default="",
        choices=accounting_methods,
        help=f"accounting method (default: '%(default)s'). Supported values: {', '.join(accounting_methods)}",
        metavar="METHOD",
        type=str,
    )
    parser.add_argument(
        "-n",
        "--allow_negative_balances",
        action="store_true",
        help="allow exchange balances to be negative (default: '%(default)s').",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        action="store",
        default="output/",
        help="Write output to OUTPUT_DIR  (default: '%(default)s')",
        metavar="OUTPUT_DIR",
        type=str,
    )
    parser.add_argument(
        "-p",
        "--prefix",
        action="store",
        default="",
        help="Prepend output file names with PREFIX",
        metavar="PREFIX",
        type=str,
    )
    parser.add_argument(
        "-t",
        "--to_date",
        action="store",
        default=MAX_DATE,
        help="Generate report up to the given date (in ISO 8601 format: e.g. YYYY-MM-DD)",
        metavar="DATE",
        type=date.fromisoformat,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"RP2 {_VERSION} (https://github.com/eprbell/rp2)",
        help="Print RP2 version",
    )
    parser.add_argument(
        "configuration_file",
        action="store",
        help="Configuration file",
        metavar="CONFIGURATION",
        type=str,
    )
    parser.add_argument(
        "input_file",
        action="store",
        help="ODS file containing input transactions",
        metavar="INPUT",
        type=str,
    )

    return parser


def _setup_paths(parser: ArgumentParser, configuration_file: str, input_file: str, output_dir: str) -> None:
    if not Path(configuration_file).exists():
        print(f"Configuration file '{configuration_file}' not found")
        parser.print_help()
        sys.exit(1)

    if not input_file.endswith(".ods"):
        print(f"Input file '{input_file}' does not end with '.ods'")
        parser.print_help()
        sys.exit(1)

    if not Path(input_file).exists():
        print(f"Input file '{input_file}' not found")
        parser.print_help()
        sys.exit(1)

    output_dir_path: Path = Path(output_dir)
    if not output_dir_path.exists():
        output_dir_path.mkdir(parents=True)
    if not output_dir_path.is_dir():
        print(f"output_dir '{output_dir}' exists but it's not a directory")
        parser.print_help()
        sys.exit(1)



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import unittest
from pathlib import Path

from abstract_test_ods_output_diff import AbstractTestODSOutputDiff, OutputPlugins

ROOT_PATH: Path = Path(os.path.dirname(__file__)).parent.absolute()


class TestLocalizedOutput(AbstractTestODSOutputDiff):  # pylint: disable=too-many-public-methods
    output_dir: Path

    @classmethod
    def setUpClass(cls) -> None:
        cls.output_dir = ROOT_PATH / Path("output") / Path(cls.__module__)

        shutil.rmtree(cls.output_dir, ignore_errors=True)

        # To test localization plumbing, we generate Japanese taxes for test_data in Kalaallisut language. Note that the localization
        # file (locales/kl/LC_MESSAGES/messages.po) doesn't contain real Kalaallisut translations, but only placeholder strings starting
        # with "__test_": this is good enough to test localization plumbing (and it would work in the same way with a real translation).
        cls._generate(cls.output_dir, test_name="test_data", config="test_data", method="fifo", generation_language="kl", country="jp")

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_test_data_rp2_full_report(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_data",
            method="fifo",
            output_plugin=OutputPlugins.RP2_FULL_REPORT,
            generation_language="kl",
            country="jp",
        )

    def test_test_data_open_positions(self) -> None:
        self._compare(
            output_dir=self.output_dir,
            test_name="test_data",
            method="fifo",
            output_plugin=OutputPlugins.OPEN_POSITIONS,
            generation_language="kl",
            country="jp",
        )


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Callable, List, Optional, cast

from rp2.abstract_entry import AbstractEntry
from rp2.abstract_transaction import AbstractTransaction
from rp2.configuration import Configuration
from rp2.in_transaction import InTransaction
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2RuntimeError, RP2TypeError, RP2ValueError


class GainLoss(AbstractEntry):
    def __init__(
        self,
        configuration: Configuration,
        crypto_amount: RP2Decimal,
        taxable_event: AbstractTransaction,
        acquired_lot: Optional[InTransaction],
    ) -> None:
        self.__taxable_event: AbstractTransaction = cast(AbstractTransaction, AbstractTransaction.type_check("taxable_event", taxable_event))
        if not taxable_event.is_taxable():
            raise RP2ValueError(f"Parameter 'taxable_event' of class {taxable_event.__class__.__name__} is not taxable: {taxable_event}")

        super().__init__(configuration, taxable_event.asset)

        self.__crypto_amount: RP2Decimal = configuration.type_check_positive_decimal("crypto_amount", crypto_amount, non_zero=True)

        if not taxable_event.transaction_type.is_earn_type():
            if acquired_lot is None:
                raise RP2TypeError("acquired_lot must not be None for non-earn-typed taxable_events")
            InTransaction.type_check("acquired_lot", acquired_lot)
        else:
            if crypto_amount != taxable_event.crypto_balance_change:
                raise RP2ValueError(
                    f"crypto_amount must be == taxable_event.crypto_balance_change for earn-typed taxable events, "
                    f"but they differ {crypto_amount} != {taxable_event.crypto_balance_change}"
                )
            if acquired_lot is not None:
                raise RP2TypeError(f"acquired_lot must be None for earn-typed taxable_events, instead it's {acquired_lot}")
        self.__acquired_lot: Optional[InTransaction] = acquired_lot

        if self.__crypto_amount > self.__taxable_event.crypto_balance_change or (self.__acquired_lot and self.__crypto_amount > self.__acquired_lot.crypto_in):
            raise RP2ValueError(
                f"crypto_amount ({self.__crypto_amount}) is greater than taxable event amount ({self.__taxable_event.crypto_balance_change}) "
                f"or acquired-lot amount ({self.__acquired_lot.crypto_in if self.__acquired_lot else 0}): {self}"
            )

        if acquired_lot is not None and taxable_event.timestamp < acquired_lot.timestamp:
            raise RP2ValueError(
                f"Timestamp {taxable_event.timestamp} of taxable_event is earlier than timestamp {acquired_lot.timestamp} " f"of acquired_lot: {self}"
            )

        if acquired_lot is not None and taxable_event.asset != acquired_lot.asset:
            raise RP2ValueError(f"taxable_event.asset ({taxable_event.asset}) != acquired_lot.asset ({acquired_lot.asset})")

    @classmethod
    def type_check(cls, name: str, instance: "AbstractEntry") -> "GainLoss":
        Configuration.type_check_parameter_name(name)
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    def __eq__(self, other: object) -> bool:
        if not other:
            return False
        if not isinstance(other, GainLoss):
            raise RP2TypeError(f"Operand has non-GainLoss value {repr(other)}")
        self_acquired_lot_internal_id: Optional[str] = self.acquired_lot.internal_id if self.acquired_lot else None
        other_acquired_lot_internal_id: Optional[str] = other.acquired_lot.internal_id if other.acquired_lot else None
        # By definition, internal_id can uniquely identify a transaction: this works even if it's the ODS line from the spreadsheet,
        # since there are no cross-asset transactions (so a spreadsheet line points to a unique transaction for that asset).
        result: bool = self.taxable_event.internal_id == other.taxable_event.internal_id and self_acquired_lot_internal_id == other_acquired_lot_internal_id
        return result

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        # By definition, internal_id can uniquely identify a transaction: this works even if it's the ODS line from the spreadsheet,
        # since there are no cross-asset transactions (so a spreadsheet line points to a unique transaction for that asset).
        return hash((self.taxable_event.internal_id, self.acquired_lot.internal_id if self.acquired_lot else None))

    def to_string(self, indent: int = 0, repr_format: bool = True, extra_data: Optional[List[str]] = None) -> str:
        self.configuration.type_check_positive_int("indent", indent)
        self.configuration.type_check_bool("repr_format", repr_format)
        if extra_data and not isinstance(extra_data, List):
            raise RP2TypeError(f"Parameter 'extra_data' is not of type List: {extra_data}")

        class_specific_data: List[str] = []
        stringify: Callable[[object], str] = str
        if repr_format:
            stringify = repr
        class_specific_data = [
            f"crypto_amount={self.crypto_amount:.8f}",
            f"fiat_cost_basis={self.fiat_cost_basis:.4f}",
            f"fiat_gain={self.fiat_gain:.4f}",
            f"is_long_term_capital_gains={stringify(self.is_long_term_capital_gains())}",
            f"taxable_event_fiat_amount_with_fee_fraction={self.taxable_event_fiat_amount_with_fee_fraction:.4f}",
            f"taxable_event_fraction_percentage={self.taxable_event_fraction_percentage:.4%}",
            f"taxable_event={self.taxable_event.to_string(indent=indent + 1, repr_format=repr_format).lstrip()}",
            f"acquired_lot_fiat_amount_with_fee_fraction={self.acquired_lot_fiat_amount_with_fee_fraction:.4f}",
            f"acquired_lot_fraction_percentage={self.acquired_lot_fraction_percentage:.4%}",
            f"acquired_lot={self.acquired_lot.to_string(indent=indent + 1, repr_format=repr_format).lstrip() if self.acquired_lot else 'None'}",
        ]
        if extra_data:
            class_specific_data.extend(extra_data)

        return super().to_string(indent=indent, repr_format=repr_format, extra_data=class_specific_data)

    @property
    def internal_id(self) -> str:
        if not self.acquired_lot:
            # earn-typed taxable event doesn't have acquired lot
            return f"{self.taxable_event.internal_id}->None"
        return f"{self.taxable_event.internal_id}->{self.acquired_lot.internal_id}"

    @property
    def timestamp(self) -> datetime:
        return self.taxable_event.timestamp

    @property
    def taxable_event(self) -> AbstractTransaction:
        return self.__taxable_event

    @property
    def acquired_lot(self) -> Optional[InTransaction]:
        return self.__acquired_lot

    @property
    def crypto_amount(self) -> RP2Decimal:
        return self.__crypto_amount

    @property
    def crypto_balance_change(self) -> RP2Decimal:
        return self.crypto_amount

    @property
    def fiat_balance_change(self) -> RP2Decimal:
        return self.taxable_event.fiat_balance_change

    @property
    def taxable_event_fiat_amount_with_fee_fraction(self) -> RP2Decimal:
        # We don't simply multiply by taxable_event_fraction_percentage to avoid potential precision loss with small percentages
        return (self.taxable_event.fiat_taxable_amount * self.crypto_amount) / self.taxable_event.crypto_balance_change

    @property
    def acquired_lot_fiat_amount_with_fee_fraction(self) -> RP2Decimal:
        if not self.acquired_lot:
            return ZERO
        # We don't simply multiply by acquired_lot_fraction_percentage to avoid potential precision loss with small percentages
        return (self.acquired_lot.fiat_in_with_fee * self.crypto_amount) / self.acquired_lot.crypto_balance_change

    @property
    def taxable_event_fraction_percentage(self) -> RP2Decimal:
        return self.crypto_amount / self.taxable_event.crypto_balance_change

    @property
    def acquired_lot_fraction_percentage(self) -> RP2Decimal:
        if not self.acquired_lot:
            # Earn-typed taxable events don't have a acquired_lot
            if not self.taxable_event.transaction_type.is_earn_type():
                raise RP2RuntimeError("Internal error: acquired lot is None but taxable event is not earn-typed")
            return ZERO
        return self.crypto_amount / self.acquired_lot.crypto_balance_change

    @property
    def fiat_cost_basis(self) -> RP2Decimal:
        if not self.acquired_lot:
            # Earn-typed taxable events don't have a acquired_lot and their cost basis is 0
            if not self.taxable_event.transaction_type.is_earn_type():
                raise RP2RuntimeError("Internal error: acquired lot is None but taxable event is not earn-typed")
            return ZERO
        # The cost basis is fiat_in + fee (as explained in https://www.irs.gov/publications/p544 and
        # https://taxbit.com/cryptocurrency-tax-guide).
        # Also note that we don't simply multiply by acquired_lot_fraction_percentage to avoid potential precision loss
        # with small percentages.
        return (self.acquired_lot.fiat_in_with_fee * self.crypto_amount) / self.acquired_lot.crypto_balance_change

    @property
    def fiat_gain(self) -> RP2Decimal:
        return self.taxable_event_fiat_amount_with_fee_fraction - self.fiat_cost_basis

    def is_long_term_capital_gains(self) -> bool:
        if not self.acquired_lot:
            # Earn-typed taxable events don't have a acquired lot and are always considered short term capital gains
            if not self.taxable_event.transaction_type.is_earn_type():
                raise RP2RuntimeError("Internal error: acquired lot is None but taxable event is not earn-typed")
            return False
        return (self.taxable_event.timestamp - self.acquired_lot.timestamp).days >= self.configuration.country.get_long_term_capital_gain_period()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, Iterator, Optional, cast

from rp2.abstract_entry import AbstractEntry
from rp2.abstract_transaction import AbstractTransaction
from rp2.accounting_engine import (
    AccountingEngine,
    AcquiredLotsExhaustedException,
    TaxableEventAndAcquiredLot,
    TaxableEventsExhaustedException,
)
from rp2.computed_data import ComputedData
from rp2.configuration import MAX_DATE, MIN_DATE, Configuration
from rp2.gain_loss import GainLoss
from rp2.gain_loss_set import GainLossSet
from rp2.in_transaction import InTransaction
from rp2.input_data import InputData
from rp2.logger import LOGGER
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2RuntimeError, RP2ValueError
from rp2.transaction_set import TransactionSet


def compute_tax(configuration: Configuration, accounting_engine: AccountingEngine, input_data: InputData) -> ComputedData:
    Configuration.type_check("configuration", configuration)
    AccountingEngine.type_check("accounting_engine", accounting_engine)
    InputData.type_check("input_data", input_data)

    unfiltered_taxable_event_set: TransactionSet = _create_unfiltered_taxable_event_set(configuration, input_data)
    LOGGER.debug("%s: Created taxable event set", input_data.asset)
    unfiltered_gain_loss_set: GainLossSet = _create_unfiltered_gain_and_loss_set(configuration, accounting_engine, input_data, unfiltered_taxable_event_set)
    LOGGER.debug("%s: Created gain-loss set", input_data.asset)

    return ComputedData(
        input_data.asset,
        unfiltered_taxable_event_set,
        unfiltered_gain_loss_set,
        input_data,
        configuration.from_date,
        configuration.to_date,
    )


def _create_unfiltered_taxable_event_set(configuration: Configuration, input_data: InputData) -> TransactionSet:
    transaction_set: TransactionSet
    entry: AbstractEntry
    transaction: AbstractTransaction
    taxable_event_set: TransactionSet = TransactionSet(configuration, "MIXED", input_data.asset, MIN_DATE, MAX_DATE)
    for transaction_set in [
        input_data.unfiltered_in_transaction_set,
        input_data.unfiltered_out_transaction_set,
        input_data.unfiltered_intra_transaction_set,
    ]:
        for entry in transaction_set:
            transaction = cast(AbstractTransaction, entry)
            if transaction.is_taxable():
                taxable_event_set.add_entry(transaction)

    return taxable_event_set


def _get_next_taxable_event_and_acquired_lot(
    accounting_engine: AccountingEngine,
    taxable_event: Optional[AbstractTransaction],
    acquired_lot: Optional[InTransaction],
    taxable_event_amount: RP2Decimal,
    acquired_lot_amount: RP2Decimal,
) -> TaxableEventAndAcquiredLot:
    new_taxable_event: AbstractTransaction
    new_acquired_lot: Optional[InTransaction]
    new_taxable_event_amount: RP2Decimal
    new_acquired_lot_amount: RP2Decimal
    (new_taxable_event, new_acquired_lot, new_taxable_event_amount, new_acquired_lot_amount) = accounting_engine.get_next_taxable_event_and_amount(
        taxable_event, acquired_lot, taxable_event_amount, acquired_lot_amount
    )
    if acquired_lot == new_acquired_lot:
        (_, new_acquired_lot, _, new_acquired_lot_amount) = accounting_engine.get_acquired_lot_for_taxable_event(
            new_taxable_event, new_acquired_lot, new_taxable_event_amount, new_acquired_lot_amount
        )
    return TaxableEventAndAcquiredLot(new_taxable_event, new_acquired_lot, new_taxable_event_amount, new_acquired_lot_amount)


def _create_unfiltered_gain_and_loss_set(
    configuration: Configuration, accounting_engine: AccountingEngine, input_data: InputData, unfiltered_taxable_event_set: TransactionSet
) -> GainLossSet:
    gain_loss_set: GainLossSet = GainLossSet(configuration, input_data.asset, MIN_DATE, MAX_DATE)
    # Create a fresh instance of accounting engine
    new_accounting_engine: AccountingEngine = accounting_engine.__class__(accounting_engine.years_2_methods)
    taxable_event_iterator: Iterator[AbstractTransaction] = iter(cast(Iterable[AbstractTransaction], unfiltered_taxable_event_set))
    acquired_lot_iterator: Iterator[InTransaction] = iter(cast(Iterable[InTransaction], input_data.unfiltered_in_transaction_set))

    new_accounting_engine.initialize(taxable_event_iterator, acquired_lot_iterator)

    try:
        gain_loss: GainLoss
        taxable_event: AbstractTransaction
        acquired_lot: Optional[InTransaction]
        taxable_event_amount: RP2Decimal
        acquired_lot_amount: RP2Decimal
        total_amount: RP2Decimal = ZERO

        # Retrieve first taxable event and acquired lot
        (taxable_event, acquired_lot, taxable_event_amount, acquired_lot_amount) = _get_next_taxable_event_and_acquired_lot(
            new_accounting_engine, None, None, ZERO, ZERO
        )

        while taxable_event:
            # Type check values returned by accounting method plugin
            AbstractTransaction.type_check("taxable_event", taxable_event)
            if acquired_lot is None:
                # There must always be at least one acquired_lot
                raise RP2RuntimeError("Parameter 'acquired_lot' is None")
            InTransaction.type_check("acquired_lot", acquired_lot)
            Configuration.type_check_positive_decimal("taxable_event_amount", taxable_event_amount)
            Configuration.type_check_positive_decimal("acquired_lot_amount", acquired_lot_amount)

            if taxable_event.transaction_type.is_earn_type():
                # Handle earn-typed transactions first: they have no acquired-lot
                gain_loss = GainLoss(configuration, taxable_event_amount, taxable_event, None)
                LOGGER.debug(
                    "tax_engine: taxable is earn: %s / %s + %s = %s: %s",
                    taxable_event_amount,
                    total_amount,
                    taxable_event_amount,
                    total_amount + taxable_event_amount,
                    gain_loss,
                )
                total_amount += taxable_event_amount
                gain_loss_set.add_entry(gain_loss)
                (taxable_event, acquired_lot, taxable_event_amount, acquired_lot_amount) = new_accounting_engine.get_next_taxable_event_and_amount(
                    taxable_event, acquired_lot, ZERO, acquired_lot_amount
                )
                continue
            if taxable_event_amount == acquired_lot_amount:
                gain_loss = GainLoss(configuration, taxable_event_amount, taxable_event, acquired_lot)
                LOGGER.debug(
                    "tax_engine: taxable == acquired: %s == %s / %s + %s = %s: %s",
                    taxable_event_amount,
                    acquired_lot_amount,
                    total_amount,
                    taxable_event_amount,
                    total_amount + taxable_event_amount,
                    gain_loss,
                )
                total_amount += taxable_event_amount
                gain_loss_set.add_entry(gain_loss)
                (taxable_event, acquired_lot, taxable_event_amount, acquired_lot_amount) = _get_next_taxable_event_and_acquired_lot(
                    new_accounting_engine, taxable_event, acquired_lot, taxable_event_amount, acquired_lot_amount
                )
            elif taxable_event_amount < acquired_lot_amount:
                gain_loss = GainLoss(configuration, taxable_event_amount, taxable_event, acquired_lot)
                LOGGER.debug(
                    "tax_engine: taxable < acquired: %s < %s / %s + %s = %s: %s",
                    taxable_event_amount,
                    acquired_lot_amount,
                    total_amount,
                    taxable_event_amount,
                    total_amount + taxable_event_amount,
                    gain_loss,
                )
                total_amount += taxable_event_amount
                gain_loss_set.add_entry(gain_loss)
                (taxable_event, acquired_lot, taxable_event_amount, acquired_lot_amount) = new_accounting_engine.get_next_taxable_event_and_amount(
                    taxable_event, acquired_lot, taxable_event_amount, acquired_lot_amount
                )
            else:  # taxable_amount > acquired_lot_amount
                gain_loss = GainLoss(configuration, acquired_lot_amount, taxable_event, acquired_lot)
                LOGGER.debug(
                    "tax_engine: taxable > acquired: %s > %s / %s + %s = %s: %s",
                    taxable_event_amount,
                    acquired_lot_amount,
                    total_amount,
                    acquired_lot_amount,
                    total_amount + acquired_lot_amount,
                    gain_loss,
                )
                total_amount += acquired_lot_amount
                gain_loss_set.add_entry(gain_loss)
                (taxable_event, acquired_lot, taxable_event_amount, acquired_lot_amount) = new_accounting_engine.get_acquired_lot_for_taxable_event(
                    taxable_event, acquired_lot, taxable_event_amount, acquired_lot_amount
                )

    except AcquiredLotsExhaustedException:
        raise RP2ValueError("Total in-transaction crypto value < total taxable crypto value") from None
    except TaxableEventsExhaustedException:
        pass

    return gain_loss_set



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest
from datetime import date
from enum import Enum
from pathlib import Path
from subprocess import run
from typing import List, Optional

from ods_diff import ods_diff

from rp2.configuration import MAX_DATE, MIN_DATE

ROOT_PATH: Path = Path(os.path.dirname(__file__)).parent.absolute()

CONFIG_PATH: Path = ROOT_PATH / Path("config")
INPUT_PATH: Path = ROOT_PATH / Path("input")
GOLDEN_PATH: Path = INPUT_PATH / Path("golden")


class OutputPlugins(Enum):
    OPEN_POSITIONS = "open_positions"
    RP2_FULL_REPORT = "rp2_full_report"
    TAX_REPORT_JP = "tax_report_jp"
    TAX_REPORT_US = "tax_report_us"


class AbstractTestODSOutputDiff(unittest.TestCase):
    # Temporarily removed lifo and hifo due to https://github.com/eprbell/rp2/issues/79
    METHODS: List[str] = ["fifo", "lifo", "hifo"]

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    @staticmethod
    def __get_time_interval(from_date: date = MIN_DATE, to_date: date = MAX_DATE) -> str:
        time_interval: str = ""
        if from_date > MIN_DATE and to_date < MAX_DATE:
            time_interval = f"{from_date}_{to_date}_"
        elif from_date > MIN_DATE and to_date >= MAX_DATE:
            time_interval = f"{from_date}_infinity_"
        elif from_date <= MIN_DATE and to_date < MAX_DATE:
            time_interval = f"0_{to_date}_"
        return time_interval

    @classmethod
    def _generate(
        cls,
        output_dir: Path,
        test_name: str,
        config: str,
        method: str,
        input_path: Path = INPUT_PATH,
        from_date: date = MIN_DATE,
        to_date: date = MAX_DATE,
        allow_negative_balances: bool = False,
        generation_language: Optional[str] = None,
        country: str = "us",
    ) -> None:
        config = test_name if config is None else config
        time_interval: str = cls.__get_time_interval(from_date, to_date)

        arguments: List[str] = [
            f"rp2_{country}",
            "-o",
            str(output_dir),
            "-p",
            f"{test_name}_{f'{country}_' if country != 'us' else ''}{f'{generation_language}_' if generation_language else ''}{time_interval}",
        ]
        if method != "mixed":
            arguments.extend(["-m", method])
        if generation_language:
            arguments.extend(["-g", generation_language])
        if from_date:
            arguments.extend(["-f", str(from_date)])
        if to_date:
            arguments.extend(["-t", str(to_date)])
        if allow_negative_balances:
            arguments.extend(["-n"])
        arguments.extend(
            [
                str(CONFIG_PATH / Path(f"{config}.ini")),
                str(input_path / Path(f"{test_name}.ods")),
            ]
        )

        run(arguments, check=True)

    def _compare(
        self,
        output_dir: Path,
        test_name: str,
        method: str,
        output_plugin: OutputPlugins,
        from_date: date = MIN_DATE,
        to_date: date = MAX_DATE,
        generation_language: Optional[str] = None,
        country: str = "us",
    ) -> None:
        time_interval: str = self.__get_time_interval(from_date, to_date)
        diff: str

        output_file_name: Path = Path(
            f"{test_name}_{f'{country}_' if country != 'us' else ''}"
            f"{f'{generation_language}_' if generation_language else ''}"
            f"{time_interval}{method}_{output_plugin.value}.ods"
        )
        full_output_file_name: Path = output_dir / output_file_name
        full_golden_file_name: Path = GOLDEN_PATH / output_file_name
        diff = ods_diff(full_golden_file_name, full_output_file_name, generate_ascii_representation=True)
        self.assertFalse(diff, msg=diff)



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import unittest

from dateutil.tz import tzutc

from rp2.configuration import Configuration
from rp2.entry_types import TransactionType
from rp2.in_transaction import InTransaction
from rp2.out_transaction import OutTransaction
from rp2.plugin.country.us import US
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError


class TestInTransaction(unittest.TestCase):
    _configuration: Configuration

    @classmethod
    def setUpClass(cls) -> None:
        TestInTransaction._configuration = Configuration("./config/test_data.ini", US())

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_transaction_type(self) -> None:
        self.assertEqual(TransactionType.AIRDROP, TransactionType.type_check_from_string("transaction_type", "airdrop"))
        self.assertEqual(TransactionType.BUY, TransactionType.type_check_from_string("transaction_type", "buy"))
        self.assertEqual(TransactionType.DONATE, TransactionType.type_check_from_string("transaction_type", "dOnAtE"))
        self.assertEqual(TransactionType.GIFT, TransactionType.type_check_from_string("transaction_type", "GIFT"))
        self.assertEqual(TransactionType.HARDFORK, TransactionType.type_check_from_string("transaction_type", "HardFork"))
        self.assertEqual(TransactionType.INTEREST, TransactionType.type_check_from_string("transaction_type", "Interest"))
        self.assertEqual(TransactionType.MINING, TransactionType.type_check_from_string("transaction_type", "MiNiNg"))
        self.assertEqual(TransactionType.MOVE, TransactionType.type_check_from_string("transaction_type", "MoVe"))
        self.assertEqual(TransactionType.SELL, TransactionType.type_check_from_string("transaction_type", "sell"))
        self.assertEqual(TransactionType.STAKING, TransactionType.type_check_from_string("transaction_type", "sTaKING"))
        self.assertEqual(TransactionType.WAGES, TransactionType.type_check_from_string("transaction_type", "WageS"))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            TransactionType.type_check_from_string(12, "buy")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'transaction_type' has non-string value .*"):
            TransactionType.type_check_from_string("transaction_type", 34.6)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'transaction_type' has non-string value .*"):
            TransactionType.type_check_from_string("transaction_type", None)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'transaction_type' has invalid transaction type value: .*"):
            TransactionType.type_check_from_string("transaction_type", "Cook")

    def test_taxable_in_transaction(self) -> None:
        in_transaction: InTransaction = InTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B1",
            "BlockFi",
            "Bob",
            "inTerest",
            RP2Decimal("1000.0"),
            RP2Decimal("2.0002"),
            fiat_fee=RP2Decimal("0"),
            fiat_in_no_fee=RP2Decimal("2000.2"),
            fiat_in_with_fee=RP2Decimal("2000.2"),
            internal_id=19,
        )

        InTransaction.type_check("my_instance", in_transaction)
        self.assertTrue(in_transaction.is_taxable())
        self.assertEqual(RP2Decimal("2000.2"), in_transaction.fiat_taxable_amount)
        self.assertEqual("19", in_transaction.internal_id)
        self.assertEqual(2021, in_transaction.timestamp.year)
        self.assertEqual(1, in_transaction.timestamp.month)
        self.assertEqual(2, in_transaction.timestamp.day)
        self.assertEqual(8, in_transaction.timestamp.hour)
        self.assertEqual(42, in_transaction.timestamp.minute)
        self.assertEqual(43, in_transaction.timestamp.second)
        self.assertEqual(882000, in_transaction.timestamp.microsecond)
        self.assertEqual(tzutc(), in_transaction.timestamp.tzinfo)
        self.assertEqual("B1", in_transaction.asset)
        self.assertEqual("BlockFi", in_transaction.exchange)
        self.assertEqual("Bob", in_transaction.holder)
        self.assertEqual(TransactionType.INTEREST, in_transaction.transaction_type)
        self.assertEqual(RP2Decimal("1000"), in_transaction.spot_price)
        self.assertEqual(RP2Decimal("2.0002"), in_transaction.crypto_in)
        self.assertEqual(RP2Decimal("2000.2"), in_transaction.fiat_in_no_fee)
        self.assertEqual(RP2Decimal("2000.2"), in_transaction.fiat_in_with_fee)
        self.assertEqual(RP2Decimal("0"), in_transaction.fiat_fee)
        self.assertEqual(RP2Decimal("2.0002"), in_transaction.crypto_balance_change)
        self.assertEqual(RP2Decimal("2000.2"), in_transaction.fiat_balance_change)

        self.assertEqual(
            str(in_transaction),
            """InTransaction:
  id=19
  timestamp=2021-01-02 08:42:43.882000 +0000
  asset=B1
  exchange=BlockFi
  holder=Bob
  transaction_type=TransactionType.INTEREST
  spot_price=1000.0000
  crypto_in=2.00020000
  fiat_fee=0.0000
  fiat_in_no_fee=2000.2000
  fiat_in_with_fee=2000.2000
  unique_id=
  is_taxable=True
  fiat_taxable_amount=2000.2000""",
        )
        self.assertEqual(
            in_transaction.to_string(2, repr_format=False, extra_data=["foobar", "qwerty"]),
            """    InTransaction:
      id=19
      timestamp=2021-01-02 08:42:43.882000 +0000
      asset=B1
      exchange=BlockFi
      holder=Bob
      transaction_type=TransactionType.INTEREST
      spot_price=1000.0000
      crypto_in=2.00020000
      fiat_fee=0.0000
      fiat_in_no_fee=2000.2000
      fiat_in_with_fee=2000.2000
      unique_id=
      is_taxable=True
      fiat_taxable_amount=2000.2000
      foobar
      qwerty""",
        )
        self.assertEqual(
            in_transaction.to_string(2, repr_format=True, extra_data=["foobar", "qwerty"]),
            (
                "    InTransaction("
                "id='19', "
                "timestamp='2021-01-02 08:42:43.882000 +0000', "
                "asset='B1', "
                "exchange='BlockFi', "
                "holder='Bob', "
                "transaction_type=<TransactionType.INTEREST: 'interest'>, "
                "spot_price=1000.0000, "
                "crypto_in=2.00020000, "
                "fiat_fee=0.0000, "
                "fiat_in_no_fee=2000.2000, "
                "fiat_in_with_fee=2000.2000, "
                "unique_id=, "
                "is_taxable=True, "
                "fiat_taxable_amount=2000.2000, "
                "foobar, "
                "qwerty)"
            ),
        )

    def test_non_taxable_in_transaction(self) -> None:
        in_transaction = InTransaction(
            self._configuration,
            "1841-01-02T15:22:03Z",
            "B2",
            "Coinbase",
            "Alice",
            "BuY",
            RP2Decimal("1000"),
            RP2Decimal("2.0002"),
            fiat_fee=RP2Decimal("20"),
            internal_id=19,
        )
        self.assertFalse(in_transaction.is_taxable())
        self.assertEqual(RP2Decimal("0"), in_transaction.fiat_taxable_amount)
        self.assertEqual("B2", in_transaction.asset)
        self.assertEqual(TransactionType.BUY, in_transaction.transaction_type)
        self.assertEqual(RP2Decimal("2.0002"), in_transaction.crypto_balance_change)
        self.assertEqual(RP2Decimal("2020.2"), in_transaction.fiat_balance_change)

        self.assertEqual(
            str(in_transaction),
            """InTransaction:
  id=19
  timestamp=1841-01-02 15:22:03.000000 +0000
  asset=B2
  exchange=Coinbase
  holder=Alice
  transaction_type=TransactionType.BUY
  spot_price=1000.0000
  crypto_in=2.00020000
  fiat_fee=20.0000
  fiat_in_no_fee=2000.2000
  fiat_in_with_fee=2020.2000
  unique_id=
  is_taxable=False
  fiat_taxable_amount=0.0000""",
        )

    def test_in_transaction_equality_and_hashing(self) -> None:
        in_transaction: InTransaction = InTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B1",
            "BlockFi",
            "Bob",
            "iNtErEsT",
            RP2Decimal("1000.0"),
            RP2Decimal("2.0002"),
            fiat_fee=RP2Decimal("0"),
            fiat_in_no_fee=RP2Decimal("2000.2"),
            fiat_in_with_fee=RP2Decimal("2000.2"),
            internal_id=19,
        )
        in_transaction2: InTransaction = InTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B1",
            "BlockFi",
            "Bob",
            "INTEReST",
            RP2Decimal("1000.0"),
            RP2Decimal("2.0002"),
            fiat_fee=RP2Decimal("0"),
            fiat_in_no_fee=RP2Decimal("2000.2"),
            fiat_in_with_fee=RP2Decimal("2000.2"),
            internal_id=19,
        )
        in_transaction3: InTransaction = InTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B1",
            "BlockFi",
            "Bob",
            "interest",
            RP2Decimal("1000.0"),
            RP2Decimal("2.0002"),
            fiat_fee=RP2Decimal("0"),
            fiat_in_no_fee=RP2Decimal("2000.2"),
            fiat_in_with_fee=RP2Decimal("2000.2"),
            internal_id=20,
        )
        self.assertEqual(in_transaction, in_transaction)
        self.assertEqual(in_transaction, in_transaction2)
        self.assertNotEqual(in_transaction, in_transaction3)
        self.assertEqual(hash(in_transaction), hash(in_transaction))
        self.assertEqual(hash(in_transaction), hash(in_transaction2))
        # These hashes would only be equal in case of hash collision (possible but very unlikely)
        self.assertNotEqual(hash(in_transaction), hash(in_transaction3))

    def test_bad_to_string(self) -> None:
        in_transaction: InTransaction = InTransaction(
            self._configuration,
            "2021-01-02T08:42:43.882Z",
            "B1",
            "BlockFi",
            "Bob",
            "INteREst",
            RP2Decimal("1000.0"),
            RP2Decimal("2.0002"),
            fiat_fee=RP2Decimal("0"),
            fiat_in_no_fee=RP2Decimal("2000.2"),
            fiat_in_with_fee=RP2Decimal("2000.2"),
            internal_id=19,
        )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'indent' has non-integer value"):
            in_transaction.to_string(None, repr_format=False, extra_data=["foobar", "qwerty"])  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'indent' has non-positive value.*"):
            in_transaction.to_string(-1, repr_format=False, extra_data=["foobar", "qwerty"])
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'repr_format' has non-bool value .*"):
            in_transaction.to_string(1, repr_format="False", extra_data=["foobar", "qwerty"])  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'extra_data' is not of type List"):
            in_transaction.to_string(1, repr_format=False, extra_data="foobar")  # type: ignore

    def test_bad_in_transaction(self) -> None:
        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string:.*"):
            InTransaction.type_check(None, None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_instance' is not of type InTransaction:.*"):
            InTransaction.type_check("my_instance", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_instance' is not of type InTransaction: OutTransaction"):
            InTransaction.type_check(
                "my_instance",
                OutTransaction(
                    self._configuration,
                    "2021-01-12T11:51:38Z",
                    "B1",
                    "BlockFi",
                    "Bob",
                    "SELL",
                    RP2Decimal("10000"),
                    RP2Decimal("1"),
                    RP2Decimal("0"),
                    internal_id=45,
                ),
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            # Bad configuration
            InTransaction(
                None,  # type: ignore
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "interest",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration' is not of type Configuration: .*"):
            # Bad configuration
            InTransaction(
                "config",  # type: ignore
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "INTEREST",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'internal_id' has non-integer .*"):
            # Bad internal_id
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "buy",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id="19",  # type: ignore
            )
        with self.assertRaisesRegex(RP2ValueError, "Error parsing parameter 'timestamp': Unknown string format: .*"):
            # Bad timestamp
            InTransaction(
                self._configuration,
                "abcdefg",
                "B1",
                "BlockFi",
                "Bob",
                "BUY",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'timestamp' value has no timezone info: .*"):
            # Bad timestamp
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43",
                "B1",
                "BlockFi",
                "Bob",
                "intErest",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'timestamp' has non-string value .*"):
            # Bad timestamp
            InTransaction(
                self._configuration,
                1111,  # type: ignore
                "B1",
                "BlockFi",
                "Bob",
                "intERest",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'asset' value is not known: .*"):
            # Bad asset
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "yyy",
                "BlockFi",
                "Bob",
                "intEResT",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'asset' has non-string value .*"):
            # Bad asset
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                1111,  # type: ignore
                "BlockFi",
                "Bob",
                "IntEResT",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'exchange' value is not known: .*"):
            # Bad exchange
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "blockfi",
                "Bob",
                "INtEResT",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'exchange' has non-string value .*"):
            # Bad exchange
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                1111,  # type: ignore
                "Bob",
                "INtEresT",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'holder' value is not known: .*"):
            # Bad holder
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "qwerty",
                "INTEresT",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'holder' has non-string value .*"):
            # Bad holder
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                1111,  # type: ignore
                "iNTEresT",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, ".*InTransaction .*, id.*invalid transaction type.*"):
            # Bad transaction type
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "seLl",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'transaction_type' has invalid transaction type value: .*"):
            # Bad transaction type
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter .* has invalid transaction type value: .*"):
            # Bad transaction type
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "cook",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter .* has non-string value .*"):
            # Bad transaction type
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                1111,  # type: ignore
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, ".*InTransaction .*, id.*parameter 'spot_price' cannot be 0"):
            # Bad spot price
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "iNTErest",
                RP2Decimal("0"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, ".*InTransaction .*, id.*parameter 'spot_price' cannot be 0"):
            # Bad spot price
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "iNTerest",
                RP2Decimal("0.00000000000001"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'spot_price' has non-positive value .*"):
            # Bad spot price
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "iNterest",
                RP2Decimal("-1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'spot_price' has non-RP2Decimal value .*"):
            # Bad spot price
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "iNtereSt",
                "1000",  # type: ignore
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'crypto_in' has zero value"):
            # Bad crypto in
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "intereSt",
                RP2Decimal("1000"),
                RP2Decimal("0"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'crypto_in' has non-positive value .*"):
            # Bad crypto in
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "interESt",
                RP2Decimal("1000"),
                RP2Decimal("-2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'crypto_in' has non-RP2Decimal value .*"):
            # Bad crypto in
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "interEst",
                RP2Decimal("1000.0"),
                "2.0002",  # type: ignore
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'fiat_fee' has non-positive value .*"):
            # Bad fiat fee
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "inteREst",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("-20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'fiat_fee' has non-RP2Decimal value .*"):
            # Bad fiat fee
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "intEREst",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee="20",  # type: ignore
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'fiat_in_no_fee' has non-positive value .*"):
            # Bad fiat in no fee
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "inTEREst",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("-2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'fiat_in_no_fee' has non-RP2Decimal value .*"):
            # Bad fiat in no fee
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "inTERESt",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee="2000.2",  # type: ignore
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'fiat_in_with_fee' has non-positive value .*"):
            # Bad fiat in with fee
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "iNTERESt",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("-2020.2"),
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'fiat_in_with_fee' has non-RP2Decimal value .*"):
            # Bad fiat in with fee
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "INTERESt",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=(1, 2, 3),  # type: ignore
                internal_id=19,
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'notes' has non-string value .*"):
            # Bad notes
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "INTEREST",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
                notes=35.6,  # type: ignore
            )
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'notes' has non-string value .*"):
            # Bad notes
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "INTeREST",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
                notes=[1, 2, 3],  # type: ignore
            )
        with self.assertRaisesRegex(RP2ValueError, "both 'crypto_fee' and 'fiat_fee' are defined: only one allowed"):
            # Bad notes
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "INTeREST",
                RP2Decimal("1000.0"),
                RP2Decimal("2.0002"),
                crypto_fee=RP2Decimal("0.02"),
                fiat_fee=RP2Decimal("20"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )

        with self.assertLogs(level="WARNING") as log:
            # Crypto in * spot price != fiat in (without fee)
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "INTerEST",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("1000"),
                fiat_in_no_fee=RP2Decimal("1900.2"),
                fiat_in_with_fee=RP2Decimal("2000.2"),
                internal_id=19,
            )
            self.assertTrue(re.search(".* InTransaction .*, id.*crypto_in.*spot_price != fiat_in_no_fee:.*", log.output[0]))

        with self.assertLogs(level="WARNING") as log:
            # fiat in (with fee) != fiat in (without fee) + fiat fee
            InTransaction(
                self._configuration,
                "2021-01-02T08:42:43.882Z",
                "B1",
                "BlockFi",
                "Bob",
                "INTerESt",
                RP2Decimal("1000"),
                RP2Decimal("2.0002"),
                fiat_fee=RP2Decimal("18"),
                fiat_in_no_fee=RP2Decimal("2000.2"),
                fiat_in_with_fee=RP2Decimal("2020.2"),
                internal_id=19,
            )
            self.assertTrue(re.search(".* InTransaction .*, id.*fiat_in_with_fee != fiat_in_no_fee.*fiat_fee:.*", log.output[0]))


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unittest
from configparser import ConfigParser
from datetime import datetime
from tempfile import NamedTemporaryFile
from typing import Optional

from dateutil.tz import tzoffset, tzutc

from rp2.abstract_country import AbstractCountry
from rp2.configuration import Configuration, Keyword
from rp2.plugin.country.us import US
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2TypeError, RP2ValueError


class TestConfiguration(unittest.TestCase):
    _country: AbstractCountry
    _configuration: Configuration

    @classmethod
    def setUpClass(cls) -> None:
        TestConfiguration._country = US()
        TestConfiguration._configuration = Configuration("./config/test_data.ini", TestConfiguration._country)

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    @staticmethod
    def _test_config(config: ConfigParser) -> Configuration:
        result: Optional[Configuration] = None
        with NamedTemporaryFile("w", delete=False) as temporary_file:
            config.write(temporary_file)
            temporary_file.flush()

            result = Configuration(temporary_file.name, TestConfiguration._country)
        os.remove(temporary_file.name)

        return result

    def test_config_file(self) -> None:
        config = ConfigParser()
        with self.assertRaisesRegex(RP2ValueError, f" no 'assets' field defined in {Keyword.GENERAL.value} section"):
            self._test_config(config)

        config[Keyword.GENERAL.value] = {}
        with self.assertRaisesRegex(RP2ValueError, f"section '{Keyword.GENERAL.value}' doesn't contain mandatory field 'assets'"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.ASSETS.value] = ""
        with self.assertRaisesRegex(RP2ValueError, f"field .* in section '{Keyword.GENERAL.value}' cannot be empty"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.ASSETS.value] = ",,,"
        with self.assertRaisesRegex(RP2ValueError, f"field .* in section '{Keyword.GENERAL.value}' cannot contain empty elements"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.ASSETS.value] = " ,, ,  ,   "
        with self.assertRaisesRegex(RP2ValueError, f"field .* in section '{Keyword.GENERAL.value}' cannot contain empty elements"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.ASSETS.value] = "B1"
        with self.assertRaisesRegex(RP2ValueError, f"section '{Keyword.GENERAL.value}' doesn't contain mandatory field .*"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.ASSETS.value] = "B1, B2"
        with self.assertRaisesRegex(RP2ValueError, f"section '{Keyword.GENERAL.value}' doesn't contain mandatory field .*"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.EXCHANGES.value] = ""
        with self.assertRaisesRegex(RP2ValueError, f"field .* in section '{Keyword.GENERAL.value}' cannot be empty"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.EXCHANGES.value] = ",,,"
        with self.assertRaisesRegex(RP2ValueError, f"field .* in section '{Keyword.GENERAL.value}' cannot contain empty elements"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.EXCHANGES.value] = ",, ,  Coinbase   ,"
        with self.assertRaisesRegex(RP2ValueError, f"field .* in section '{Keyword.GENERAL.value}' cannot contain empty elements"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.EXCHANGES.value] = "Coinbase"
        with self.assertRaisesRegex(RP2ValueError, f"section '{Keyword.GENERAL.value}' doesn't contain mandatory field .*"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.EXCHANGES.value] = "BlockFi,Coinbase,Kraken"
        with self.assertRaisesRegex(RP2ValueError, f"section '{Keyword.GENERAL.value}' doesn't contain mandatory field .*"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.HOLDERS.value] = ""
        with self.assertRaisesRegex(RP2ValueError, f"field .* in section '{Keyword.GENERAL.value}' cannot be empty"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.HOLDERS.value] = ",,,"
        with self.assertRaisesRegex(RP2ValueError, f"field .* in section '{Keyword.GENERAL.value}' cannot contain empty elements"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.HOLDERS.value] = "Bob,, ,  ,   "
        with self.assertRaisesRegex(RP2ValueError, f"field .* in section '{Keyword.GENERAL.value}' cannot contain empty elements"):
            self._test_config(config)

        config[Keyword.GENERAL.value][Keyword.HOLDERS.value] = "Bob"
        with self.assertRaisesRegex(RP2ValueError, f"empty '{Keyword.IN_HEADER.value}' section"):
            self._test_config(config)

        config[Keyword.IN_HEADER.value] = {}
        with self.assertRaisesRegex(RP2ValueError, f"section '{Keyword.IN_HEADER.value}' cannot be empty"):
            self._test_config(config)

        config[Keyword.IN_HEADER.value][Keyword.TIMESTAMP.value] = "abc"
        with self.assertRaisesRegex(RP2ValueError, f"invalid column value for field .* in section '{Keyword.IN_HEADER.value}' .*integer was expected.*:"):
            self._test_config(config)

        config[Keyword.IN_HEADER.value][Keyword.TIMESTAMP.value] = ""
        with self.assertRaisesRegex(RP2ValueError, f"invalid column value for field .* in section '{Keyword.IN_HEADER.value}' .*integer was expected.*:"):
            self._test_config(config)

        config[Keyword.IN_HEADER.value][Keyword.TIMESTAMP.value] = "8.8"
        with self.assertRaisesRegex(RP2ValueError, f"invalid column value for field .* in section '{Keyword.IN_HEADER.value}' .*integer was expected.*:"):
            self._test_config(config)

        config[Keyword.IN_HEADER.value][Keyword.TIMESTAMP.value] = "-14"
        with self.assertRaisesRegex(
            RP2ValueError, f"invalid column value for field .* in section '{Keyword.IN_HEADER.value}' .*positive integer was expected.*:"
        ):
            self._test_config(config)

        config[Keyword.IN_HEADER.value][Keyword.TIMESTAMP.value] = "0"
        config[Keyword.IN_HEADER.value]["foobar"] = "66"
        with self.assertRaisesRegex(RP2ValueError, f"invalid column header in section '{Keyword.IN_HEADER.value}': .*"):
            self._test_config(config)

        config[Keyword.IN_HEADER.value] = {
            Keyword.TIMESTAMP.value: "0",
            Keyword.ASSET.value: "6",
            Keyword.EXCHANGE.value: "1",
            Keyword.HOLDER.value: "2",
            Keyword.TRANSACTION_TYPE.value: "5",
            Keyword.SPOT_PRICE.value: "8",
            Keyword.CRYPTO_IN.value: "7",
            Keyword.FIAT_FEE.value: "11",
            Keyword.FIAT_IN_NO_FEE.value: "9",
            Keyword.FIAT_IN_WITH_FEE.value: "10",
            Keyword.NOTES.value: "12",
        }

        config[Keyword.OUT_HEADER.value] = {}
        with self.assertRaisesRegex(RP2ValueError, f"section '{Keyword.OUT_HEADER.value}' cannot be empty"):
            self._test_config(config)

        config[Keyword.OUT_HEADER.value][Keyword.ASSET.value] = "def ghi"
        with self.assertRaisesRegex(RP2ValueError, f"invalid column value for field .* in section '{Keyword.OUT_HEADER.value}' .*integer was expected.*:"):
            self._test_config(config)

        config[Keyword.OUT_HEADER.value][Keyword.ASSET.value] = ""
        with self.assertRaisesRegex(RP2ValueError, f"invalid column value for field .* in section '{Keyword.OUT_HEADER.value}' .*integer was expected.*:"):
            self._test_config(config)

        config[Keyword.OUT_HEADER.value][Keyword.ASSET.value] = "147.222"
        with self.assertRaisesRegex(RP2ValueError, f"invalid column value for field .* in section '{Keyword.OUT_HEADER.value}' .*integer was expected.*:"):
            self._test_config(config)

        config[Keyword.OUT_HEADER.value][Keyword.ASSET.value] = "-903"
        with self.assertRaisesRegex(
            RP2ValueError, f"invalid column value for field .* in section '{Keyword.OUT_HEADER.value}' .*positive integer was expected.*:"
        ):
            self._test_config(config)

        config[Keyword.OUT_HEADER.value][Keyword.ASSET.value] = "0"
        config[Keyword.OUT_HEADER.value]["__dummy"] = "718"
        with self.assertRaisesRegex(RP2ValueError, f"invalid column header in section '{Keyword.OUT_HEADER.value}': .*"):
            self._test_config(config)

        config[Keyword.OUT_HEADER.value] = {
            Keyword.TIMESTAMP.value: "0",
            Keyword.ASSET.value: "6",
            Keyword.EXCHANGE.value: "1",
            Keyword.HOLDER.value: "2",
            Keyword.TRANSACTION_TYPE.value: "5",
            Keyword.SPOT_PRICE.value: "8",
            Keyword.CRYPTO_OUT_NO_FEE.value: "7",
            Keyword.CRYPTO_FEE.value: "9",
            Keyword.NOTES.value: "12",
        }
        with self.assertRaisesRegex(RP2ValueError, f"empty '{Keyword.INTRA_HEADER.value}' section"):
            self._test_config(config)

        config[Keyword.INTRA_HEADER.value] = {}
        with self.assertRaisesRegex(RP2ValueError, f"section '{Keyword.INTRA_HEADER.value}' cannot be empty"):
            self._test_config(config)

        config[Keyword.INTRA_HEADER.value][Keyword.FROM_EXCHANGE.value] = "x y z"
        with self.assertRaisesRegex(RP2ValueError, f"invalid column value for field .* in section '{Keyword.INTRA_HEADER.value}' .*integer was expected.*:"):
            self._test_config(config)

        config[Keyword.INTRA_HEADER.value][Keyword.FROM_EXCHANGE.value] = ""
        with self.assertRaisesRegex(RP2ValueError, f"invalid column value for field .* in section '{Keyword.INTRA_HEADER.value}' .*integer was expected.*:"):
            self._test_config(config)

        config[Keyword.INTRA_HEADER.value][Keyword.FROM_EXCHANGE.value] = "52.76"
        with self.assertRaisesRegex(RP2ValueError, f"invalid column value for field .* in section '{Keyword.INTRA_HEADER.value}' .*integer was expected.*:"):
            self._test_config(config)

        config[Keyword.INTRA_HEADER.value][Keyword.FROM_EXCHANGE.value] = "-55082824"
        with self.assertRaisesRegex(
            RP2ValueError, f"invalid column value for field .* in section '{Keyword.INTRA_HEADER.value}' .*positive integer was expected.*:"
        ):
            self._test_config(config)

        config[Keyword.INTRA_HEADER.value][Keyword.FROM_EXCHANGE.value] = "0"
        config[Keyword.INTRA_HEADER.value]["whatever"] = "349"
        with self.assertRaisesRegex(RP2ValueError, f"invalid column header in section '{Keyword.INTRA_HEADER.value}': .*"):
            self._test_config(config)

        config[Keyword.INTRA_HEADER.value] = {
            Keyword.TIMESTAMP.value: "0",
            Keyword.ASSET.value: "6",
            Keyword.FROM_EXCHANGE.value: "1",
            Keyword.FROM_HOLDER.value: "2",
            Keyword.TO_EXCHANGE.value: "3",
            Keyword.TO_HOLDER.value: "4",
            Keyword.SPOT_PRICE.value: "8",
            Keyword.CRYPTO_SENT.value: "7",
            Keyword.CRYPTO_RECEIVED.value: "10",
            Keyword.NOTES.value: "12",
        }
        self._test_config(config)

        config[Keyword.IN_HEADER.value][Keyword.TIMESTAMP.value] = "8"
        with self.assertRaisesRegex(RP2ValueError, f"fields .* and .* have the same value in section '{Keyword.IN_HEADER.value}': .*"):
            self._test_config(config)
        config[Keyword.IN_HEADER.value][Keyword.TIMESTAMP.value] = "0"

        config[Keyword.OUT_HEADER.value][Keyword.ASSET.value] = "12"
        with self.assertRaisesRegex(RP2ValueError, f"fields .* and .* have the same value in section '{Keyword.OUT_HEADER.value}': .*"):
            self._test_config(config)
        config[Keyword.OUT_HEADER.value][Keyword.ASSET.value] = "6"

        config[Keyword.INTRA_HEADER.value][Keyword.FROM_EXCHANGE.value] = "4"
        with self.assertRaisesRegex(RP2ValueError, f"fields .* and .* have the same value in section '{Keyword.INTRA_HEADER.value}': .*"):
            self._test_config(config)
        config[Keyword.INTRA_HEADER.value][Keyword.FROM_EXCHANGE.value] = "1"

        config[Keyword.GENERAL.value][Keyword.ASSETS.value] = "B1, B2, B1"
        with self.assertRaisesRegex(RP2ValueError, f"field .* in section '{Keyword.GENERAL.value}' contains duplicate elements"):
            self._test_config(config)
        config[Keyword.GENERAL.value][Keyword.ASSETS.value] = "B1, B2"

        config[Keyword.GENERAL.value][Keyword.EXCHANGES.value] = "BlockFi,BlockFi"
        with self.assertRaisesRegex(RP2ValueError, f"field .* in section '{Keyword.GENERAL.value}' contains duplicate elements"):
            self._test_config(config)
        config[Keyword.GENERAL.value][Keyword.EXCHANGES.value] = "BlockFi,Coinbase,Kraken"

        config[Keyword.GENERAL.value][Keyword.HOLDERS.value] = "Bob, Alice, Alice"
        with self.assertRaisesRegex(RP2ValueError, f"field .* in section '{Keyword.GENERAL.value}' contains duplicate elements"):
            self._test_config(config)
        config[Keyword.GENERAL.value][Keyword.HOLDERS.value] = "Bob"

        config[f"{Keyword.GENERAL.value} foo bar"] = config[Keyword.GENERAL.value]
        with self.assertRaisesRegex(RP2ValueError, f"section '{Keyword.GENERAL.value}' found multiple times in configuration file"):
            self._test_config(config)
        del config[f"{Keyword.GENERAL.value} foo bar"]

        config[f"{Keyword.IN_HEADER.value} yy"] = config[Keyword.IN_HEADER.value]
        with self.assertRaisesRegex(RP2ValueError, f"section '{Keyword.IN_HEADER.value}' found multiple times in configuration file"):
            self._test_config(config)
        del config[f"{Keyword.IN_HEADER.value} yy"]

        config[f"{Keyword.OUT_HEADER.value} 1 2 3"] = config[Keyword.OUT_HEADER.value]
        with self.assertRaisesRegex(RP2ValueError, f"section '{Keyword.OUT_HEADER.value}' found multiple times in configuration file"):
            self._test_config(config)
        del config[f"{Keyword.OUT_HEADER.value} 1 2 3"]

        config[f"{Keyword.INTRA_HEADER.value} "] = config[Keyword.INTRA_HEADER.value]
        with self.assertRaisesRegex(RP2ValueError, f"section '{Keyword.INTRA_HEADER.value}' found multiple times in configuration file"):
            self._test_config(config)
        del config[f"{Keyword.INTRA_HEADER.value} "]

        config["__bad_section"] = config[Keyword.IN_HEADER.value]
        with self.assertRaisesRegex(RP2ValueError, "invalid section '__bad_section' found"):
            self._test_config(config)
        del config["__bad_section"]

        config[Keyword.ACCOUNTING_METHODS.value] = {}
        with self.assertRaisesRegex(RP2ValueError, f"section '{Keyword.ACCOUNTING_METHODS.value}' cannot be empty"):
            self._test_config(config)

        config[Keyword.ACCOUNTING_METHODS.value] = {"1218": "fifo"}
        with self.assertRaisesRegex(RP2ValueError, "invalid year value in accounting method section .*integer > 1970 was expected"):
            self._test_config(config)

        config[Keyword.ACCOUNTING_METHODS.value] = {"abc": "fifo"}
        with self.assertRaisesRegex(RP2ValueError, "invalid year value in accounting method section .*integer was expected"):
            self._test_config(config)

        config[Keyword.ACCOUNTING_METHODS.value] = {"1970": "hifo", "2020": "lifo"}

    def test_creation(self) -> None:
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'country' is not of type AbstractCountry: .*"):
            Configuration("./config/test_data.ini", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration_path' has non-string value .*"):
            Configuration(None, self._country)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'configuration_path' has non-string value .*"):
            Configuration(111, self._country)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "/non/existing/file does not exist"):
            Configuration("/non/existing/file", self._country)
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'from_date' is not of type date"):
            Configuration("./config/test_data.ini", self._country, from_date=None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'from_date' is not of type date"):
            Configuration("./config/test_data.ini", self._country, from_date="foobar")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'to_date' is not of type date"):
            Configuration("./config/test_data.ini", self._country, to_date=None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'to_date' is not of type date"):
            Configuration("./config/test_data.ini", self._country, to_date="foobar")  # type: ignore

    def test_argument_packs(self) -> None:
        self.assertEqual(
            str(self._configuration.get_in_table_constructor_argument_pack([0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110, 120])),
            (
                "{'timestamp': 0, 'asset': 60, 'exchange': 10, 'holder': 20, 'transaction_type': 50, 'spot_price': 80, "
                "'crypto_in': 70, 'fiat_fee': 110, 'fiat_in_no_fee': 90, 'fiat_in_with_fee': 100, 'notes': 120}"
            ),
        )
        self.assertEqual(
            str(self._configuration.get_out_table_constructor_argument_pack([0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110, 120])),
            (
                "{'timestamp': 0, 'asset': 60, 'exchange': 10, 'holder': 20, 'transaction_type': 50, 'spot_price': 80, "
                "'crypto_out_no_fee': 70, 'crypto_fee': 90, 'notes': 120}"
            ),
        )
        self.assertEqual(
            str(self._configuration.get_intra_table_constructor_argument_pack([0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110, 120])),
            (
                "{'timestamp': 0, 'asset': 60, 'from_exchange': 10, 'from_holder': 20, 'to_exchange': 30, 'to_holder': 40, "
                "'spot_price': 80, 'crypto_sent': 70, 'crypto_received': 100, 'notes': 120}"
            ),
        )

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'data' value is not a List"):
            self._configuration.get_in_table_constructor_argument_pack(None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'data' value is not a List"):
            self._configuration.get_out_table_constructor_argument_pack("foobar")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'data' value is not a List"):
            self._configuration.get_intra_table_constructor_argument_pack(8)  # type: ignore

        with self.assertRaisesRegex(RP2ValueError, "Parameter 'data' has length .*, but required minimum from in-table headers in .* is .*"):
            self._configuration.get_in_table_constructor_argument_pack([1, 2, 3])
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'data' has length .*, but required minimum from out-table headers in .* is .*"):
            self._configuration.get_out_table_constructor_argument_pack([])
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'data' has length .*, but required minimum from intra-table headers in .* is .*"):
            self._configuration.get_intra_table_constructor_argument_pack([1])

    def test_get_column_position(self) -> None:
        self.assertEqual(self._configuration.get_in_table_column_position("timestamp"), 0)
        self.assertEqual(self._configuration.get_in_table_column_position("asset"), 6)
        self.assertEqual(self._configuration.get_in_table_column_position("exchange"), 1)
        self.assertEqual(self._configuration.get_in_table_column_position("holder"), 2)
        self.assertEqual(self._configuration.get_in_table_column_position("transaction_type"), 5)
        self.assertEqual(self._configuration.get_in_table_column_position("spot_price"), 8)
        self.assertEqual(self._configuration.get_in_table_column_position("crypto_in"), 7)
        self.assertEqual(self._configuration.get_in_table_column_position("fiat_fee"), 11)
        self.assertEqual(self._configuration.get_in_table_column_position("fiat_in_no_fee"), 9)
        self.assertEqual(self._configuration.get_in_table_column_position("fiat_in_with_fee"), 10)
        self.assertEqual(self._configuration.get_in_table_column_position("notes"), 12)

        self.assertEqual(self._configuration.get_out_table_column_position("timestamp"), 0)
        self.assertEqual(self._configuration.get_out_table_column_position("asset"), 6)
        self.assertEqual(self._configuration.get_out_table_column_position("exchange"), 1)
        self.assertEqual(self._configuration.get_out_table_column_position("holder"), 2)
        self.assertEqual(self._configuration.get_out_table_column_position("transaction_type"), 5)
        self.assertEqual(self._configuration.get_out_table_column_position("spot_price"), 8)
        self.assertEqual(self._configuration.get_out_table_column_position("crypto_out_no_fee"), 7)
        self.assertEqual(self._configuration.get_out_table_column_position("crypto_fee"), 9)
        self.assertEqual(self._configuration.get_out_table_column_position("notes"), 12)

        self.assertEqual(self._configuration.get_intra_table_column_position("timestamp"), 0)
        self.assertEqual(self._configuration.get_intra_table_column_position("asset"), 6)
        self.assertEqual(self._configuration.get_intra_table_column_position("from_exchange"), 1)
        self.assertEqual(self._configuration.get_intra_table_column_position("from_holder"), 2)
        self.assertEqual(self._configuration.get_intra_table_column_position("to_exchange"), 3)
        self.assertEqual(self._configuration.get_intra_table_column_position("to_holder"), 4)
        self.assertEqual(self._configuration.get_intra_table_column_position("spot_price"), 8)
        self.assertEqual(self._configuration.get_intra_table_column_position("crypto_sent"), 7)
        self.assertEqual(self._configuration.get_intra_table_column_position("crypto_received"), 10)
        self.assertEqual(self._configuration.get_intra_table_column_position("notes"), 12)

        with self.assertRaisesRegex(RP2TypeError, "Parameter 'input_parameter' has non-string value .*"):
            self._configuration.get_in_table_column_position(None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'input_parameter' has non-string value .*"):
            self._configuration.get_out_table_column_position(12)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'input_parameter' has non-string value .*"):
            self._configuration.get_intra_table_column_position([1, 2])  # type: ignore

        with self.assertRaisesRegex(RP2ValueError, "Unknown 'input_parameter' .*"):
            self._configuration.get_in_table_column_position("non_existent")
        with self.assertRaisesRegex(RP2ValueError, "Unknown 'input_parameter' .*"):
            self._configuration.get_out_table_column_position("foobar")
        with self.assertRaisesRegex(RP2ValueError, "Unknown 'input_parameter' .*"):
            self._configuration.get_intra_table_column_position("whatever")

    def test_string_representation(self) -> None:
        self.assertEqual(
            str(self._configuration),
            (
                "Configuration(configuration_path=./config/test_data.ini, country=US(country_iso_code=us, "
                "currency_iso_code=usd, long_term_capital_gain_period=365), from_date=non-specified, "
                "to_date=non-specified, in_header={'timestamp': 0, 'asset': 6, 'exchange': 1, 'holder': 2, 'transaction_type': 5, 'spot_price': 8, "
                "'crypto_in': 7, 'fiat_fee': 11, 'fiat_in_no_fee': 9, 'fiat_in_with_fee': 10, 'notes': 12}, out_header={'timestamp': 0, 'asset': 6, "
                "'exchange': 1, 'holder': 2, 'transaction_type': 5, 'spot_price': 8, 'crypto_out_no_fee': 7, 'crypto_fee': 9, 'notes': 12}, "
                "intra_header={'timestamp': 0, 'asset': 6, 'from_exchange': 1, 'from_holder': 2, 'to_exchange': 3, 'to_holder': 4, "
                "'spot_price': 8, 'crypto_sent': 7, 'crypto_received': 10, 'notes': 12}, assets=['B1', 'B2', 'B3', 'B4'], "
                "exchanges=['BlockFi', 'Coinbase', 'Coinbase Pro', 'Kraken'], holders=['Alice', 'Bob'])"
            ),
        )

    def test_internal_id(self) -> None:
        self.assertEqual(13, self._configuration.type_check_internal_id("internal_id", 13))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_internal_id(None, 10)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'internal_id' has non-integer value .*"):
            self._configuration.type_check_internal_id("internal_id", 7.7)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'internal_id' has non-integer value .*"):
            self._configuration.type_check_internal_id("internal_id", None)  # type: ignore

    def test_timestamp(self) -> None:
        date: datetime = self._configuration.type_check_timestamp_from_string("timestamp", "2020-06-21T23:29:03.117Z")
        self.assertEqual(2020, date.year)
        self.assertEqual(6, date.month)
        self.assertEqual(21, date.day)
        self.assertEqual(23, date.hour)
        self.assertEqual(29, date.minute)
        self.assertEqual(3, date.second)
        self.assertEqual(117000, date.microsecond)
        self.assertEqual(tzutc(), date.tzinfo)

        date = self._configuration.type_check_timestamp_from_string("timestamp", "2020-12-01 03:59:49 -04:00")
        self.assertEqual(2020, date.year)
        self.assertEqual(12, date.month)
        self.assertEqual(1, date.day)
        self.assertEqual(3, date.hour)
        self.assertEqual(59, date.minute)
        self.assertEqual(49, date.second)
        self.assertEqual(0, date.microsecond)
        self.assertEqual(tzoffset(None, -14400), date.tzinfo)

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_timestamp_from_string(12, "2020-12-01 03:59:49 -04:00")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'timestamp' has non-string value .*"):
            self._configuration.type_check_timestamp_from_string("timestamp", 2020)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'timestamp' has non-string value .*"):
            self._configuration.type_check_timestamp_from_string("timestamp", None)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'timestamp' value has no timezone info: .*"):
            self._configuration.type_check_timestamp_from_string("timestamp", "2020-12-01 03:59:49")
        with self.assertRaisesRegex(RP2ValueError, "hour must be in 0..23"):
            self._configuration.type_check_timestamp_from_string("timestamp", "2020-12-01 25:59:49")
        with self.assertRaisesRegex(RP2ValueError, "Unknown string format: .*"):
            self._configuration.type_check_timestamp_from_string("timestamp", "foo bar baz")

    def test_exchange(self) -> None:
        self.assertEqual("Coinbase Pro", self._configuration.type_check_exchange("exchange", "Coinbase Pro"))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_exchange(None, "Coinbase")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'exchange' has non-string value .*"):
            self._configuration.type_check_exchange("exchange", 34.6)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'exchange' has non-string value .*"):
            self._configuration.type_check_exchange("exchange", None)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'exchange' value is not known: .*"):
            self._configuration.type_check_exchange("exchange", "coinbase pro")
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'exchange' value is not known: .*"):
            self._configuration.type_check_exchange("exchange", "Coinbase Proo")

    def test_holder(self) -> None:
        self.assertEqual("Bob", self._configuration.type_check_holder("holder", "Bob"))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_holder(None, "Alice")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'holder' has non-string value .*"):
            self._configuration.type_check_holder("holder", 34.6)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'holder' has non-string value .*"):
            self._configuration.type_check_holder("holder", None)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'holder' value is not known: .*"):
            self._configuration.type_check_holder("holder", "John")

    def test_asset(self) -> None:
        self.assertEqual("B1", self._configuration.type_check_asset("asset", "B1"))
        self.assertEqual("B2", self._configuration.type_check_asset("asset", "B2"))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_asset(None, "B1")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'asset' has non-string value .*"):
            self._configuration.type_check_asset("asset", 34.6)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'asset' has non-string value .*"):
            self._configuration.type_check_asset("asset", None)  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'asset' value is not known: .*"):
            self._configuration.type_check_asset("asset", "btc")
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'asset' value is not known: .*"):
            self._configuration.type_check_asset("asset", "qwerty")

    def test_string(self) -> None:
        self.assertEqual("foobar", self._configuration.type_check_string("my_string", "foobar"))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_string(None, "foobar")  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_string' has non-string value .*"):
            self._configuration.type_check_string("my_string", [1, 2, 3])  # type: ignore

    def test_int(self) -> None:
        self.assertEqual(5, self._configuration.type_check_int("my_int", 5))
        self.assertEqual(5, self._configuration.type_check_positive_int("my_int", 5))
        self.assertEqual(5, self._configuration.type_check_positive_int("my_int", 5, non_zero=True))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_int(None, 5)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_int' has non-integer value .*"):
            self._configuration.type_check_int("my_int", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_int' has non-integer value .*"):
            self._configuration.type_check_int("my_int", "5")  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_positive_int(None, 5)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_int' has non-integer value .*"):
            self._configuration.type_check_positive_int("my_int", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_int' has non-integer value .*"):
            self._configuration.type_check_positive_int("my_int", "5")  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'my_int' has non-positive value .*"):
            self._configuration.type_check_positive_int("my_int", -5)
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'my_int' has zero value"):
            self._configuration.type_check_positive_int("my_int", 0, non_zero=True)

    def test_float(self) -> None:
        self.assertEqual(5.5, self._configuration.type_check_float("my_float", 5.5))
        self.assertEqual(5.5, self._configuration.type_check_positive_float("my_float", 5.5))
        self.assertEqual(5.5, self._configuration.type_check_positive_float("my_float", 5.5, non_zero=True))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_float(None, 5.5)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_float' has non-numeric value .*"):
            self._configuration.type_check_float("my_float", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_float' has non-numeric value .*"):
            self._configuration.type_check_float("my_float", "5.5")  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_positive_float(None, 5.5)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_float' has non-numeric value .*"):
            self._configuration.type_check_positive_float("my_float", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_float' has non-numeric value .*"):
            self._configuration.type_check_positive_float("my_float", "5.5")  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'my_float' has non-positive value .*"):
            self._configuration.type_check_positive_float("my_float", -5.5)
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'my_float' has zero value"):
            self._configuration.type_check_positive_float("my_float", 0, non_zero=True)

    def test_rp2_decimal(self) -> None:
        minus_one: RP2Decimal = RP2Decimal("-1")
        one: RP2Decimal = RP2Decimal("1")

        self.assertEqual(minus_one, self._configuration.type_check_decimal("minus_one", minus_one))
        self.assertEqual(one, self._configuration.type_check_positive_decimal("one", one))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_decimal(None, one)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_decimal' has non-RP2Decimal value .*"):
            self._configuration.type_check_decimal("my_decimal", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_decimal' has non-RP2Decimal value .*"):
            self._configuration.type_check_decimal("my_decimal", "5.5")  # type: ignore

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_positive_decimal(None, one)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_decimal' has non-RP2Decimal value .*"):
            self._configuration.type_check_positive_decimal("my_decimal", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_decimal' has non-RP2Decimal value .*"):
            self._configuration.type_check_positive_decimal("my_decimal", "5.5")  # type: ignore
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'my_decimal' has non-positive value .*"):
            self._configuration.type_check_positive_decimal("my_decimal", minus_one)
        with self.assertRaisesRegex(RP2ValueError, "Parameter 'my_decimal' has zero value"):
            self._configuration.type_check_positive_decimal("my_decimal", ZERO, non_zero=True)

    def test_bool(self) -> None:
        self.assertEqual(True, self._configuration.type_check_bool("my_bool", True))
        self.assertEqual(False, self._configuration.type_check_bool("my_bool", False))

        with self.assertRaisesRegex(RP2TypeError, "Parameter name is not a string: .*"):
            self._configuration.type_check_bool(14, True)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_bool' has non-bool value .*"):
            self._configuration.type_check_bool("my_bool", None)  # type: ignore
        with self.assertRaisesRegex(RP2TypeError, "Parameter 'my_bool' has non-bool value .*"):
            self._configuration.type_check_bool("my_bool", "True")  # type: ignore


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import unittest
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, List

import ezodf
from abstract_test_ods_output_diff import AbstractTestODSOutputDiff, OutputPlugins
from dateutil.tz import gettz

from rp2.rp2_error import RP2RuntimeError

ROOT_PATH: Path = Path(os.path.dirname(__file__)).parent.absolute()


class TestLargeInput(AbstractTestODSOutputDiff):
    output_dir: Path

    @classmethod
    def setUpClass(cls) -> None:
        cls.output_dir = ROOT_PATH / Path("output") / Path(cls.__module__)

        shutil.rmtree(cls.output_dir, ignore_errors=True)

        cls._generate_large_input(cls.output_dir)
        for method in AbstractTestODSOutputDiff.METHODS:
            AbstractTestODSOutputDiff._generate(
                cls.output_dir, test_name="test_large_input", config="test_large_input", method=method, input_path=cls.output_dir
            )

    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    @staticmethod
    def _fill_row(sheet: Any, row_index: int, row: List[Any]) -> int:
        for col_index, element in enumerate(row):
            sheet[row_index, col_index].set_value(element)
        return row_index + 1

    @classmethod
    def _generate_large_input(cls, output_dir: Path) -> None:
        output_file_path: Path = output_dir / Path("test_large_input.ods")
        if not output_dir.exists():
            output_dir.mkdir(parents=True)
        if not output_dir.is_dir():
            raise RP2RuntimeError(f"output_dir '{str(output_dir)}' exists but it's not a directory")

        output_file: Any = ezodf.newdoc("ods", str(output_file_path), template=None)
        row_index: int = 0
        timestamp: datetime = datetime(2015, 1, 1, 1, 1, 1, 0, tzinfo=gettz("UTC"))
        asset: str = "B1"
        exchanges: List[str] = ["Coinbase", "Kraken"]
        holders: List[str] = ["Alice", "Bob"]
        in_type: List[str] = ["buy", "interest"]
        out_type: List[str] = ["sell", "donate"]

        column_total: int = 30
        table_row_total: int = 200 if "RP2_TEST_TABLE_SIZE" not in os.environ else int(str(os.environ.get("RP2_TEST_TABLE_SIZE")))

        sheet: Any = ezodf.Table("B1")
        output_file.sheets += sheet

        sheet.reset(size=(table_row_total * 2 + 100, column_total))

        row_index = cls._fill_row(sheet, row_index, ["IN"])
        row_index = cls._fill_row(
            sheet,
            row_index,
            [
                "timestamp",
                "asset",
                "exchange",
                "holder",
                "type",
                "spot",
                "crypto in",
                "fiat fee",
                "notes",
            ],
        )
        for i in range(0, table_row_total):
            row_index = cls._fill_row(
                sheet,
                row_index,
                [
                    timestamp,
                    asset,
                    exchanges[int(i / 5) % 2],
                    holders[int(i / 10) % 2],
                    in_type[int(i / 10) % 2],
                    float(1 + i % 10) / 10,
                    1 + i % 5,
                    0,
                ],
            )
            if i < table_row_total * 0.93:
                timestamp += timedelta(minutes=1)
        row_index = cls._fill_row(sheet, row_index, ["TABLE END"])

        row_index = cls._fill_row(sheet, row_index, [""])

        row_index = cls._fill_row(sheet, row_index, ["OUT"])
        row_index = cls._fill_row(
            sheet,
            row_index,
            [
                "timestamp",
                "asset",
                "exchange",
                "holder",
                "type",
                "spot",
                "crypto out",
                "crypto fee",
                "notes",
            ],
        )
        for i in range(0, table_row_total):
            row_index = cls._fill_row(
                sheet,
                row_index,
                [
                    timestamp,
                    asset,
                    exchanges[int(i / 5) % 2],
                    holders[int(i / 10) % 2],
                    out_type[int(i / 10) % 2],
                    float(1 + i % 10) / 10,
                    float(1 + i % 10) / 20,
                    0,
                ],
            )
            if i < table_row_total * 0.80:
                timestamp += timedelta(minutes=1)
        row_index = cls._fill_row(sheet, row_index, ["TABLE END"])

        output_file.save()

    def test_large_input_rp2_full_report(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_large_input", method=method, output_plugin=OutputPlugins.RP2_FULL_REPORT)

    def test_large_input_tax_report_us(self) -> None:
        for method in self.METHODS:
            self._compare(output_dir=self.output_dir, test_name="test_large_input", method=method, output_plugin=OutputPlugins.TAX_REPORT_US)


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 Larision
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Set

from rp2.abstract_country import AbstractCountry
from rp2.rp2_main import rp2_main


# ES-specific class
class ES(AbstractCountry):
    def __init__(self) -> None:
        super().__init__("es", "eur")

    # Measured in days
    def get_long_term_capital_gain_period(self) -> int:
        return 365

    # Default accounting method to use if the user doesn't specify one on the command line
    def get_default_accounting_method(self) -> str:
        return "fifo"

    # Set of accounting methods accepted in the country
    def get_accounting_methods(self) -> Set[str]:
        return {"fifo"}

    # Default set of generators to use if the user doesn't specify them on the command line
    def get_report_generators(self) -> Set[str]:
        return {
            "open_positions",
            "rp2_full_report",
        }

    # Default language to use at report generation if the user doesn't specify it on the command line (in ISO 639-1 format)
    def get_default_generation_language(self) -> str:
        return "es"


# ES-specific entry point
def rp2_entry() -> None:
    rp2_main(ES())



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from datetime import date
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Set, Tuple, cast

from rp2.abstract_country import AbstractCountry
from rp2.computed_data import ComputedData
from rp2.entry_types import TransactionType
from rp2.gain_loss import GainLoss
from rp2.gain_loss_set import GainLossSet
from rp2.logger import create_logger
from rp2.plugin.report.abstract_ods_generator import AbstractODSGenerator
from rp2.rp2_error import RP2TypeError

LOGGER: logging.Logger = create_logger("tax_report_us")


class SheetNames(Enum):
    AIRDROPS: str = "Airdrops"
    CAPITAL_GAINS: str = "Capital Gains"
    DONATIONS: str = "Donations"
    GIFTS: str = "Gifts"
    HARDFORKS: str = "Hard Forks"
    INCOME: str = "Income"
    INTEREST: str = "Interest"
    INVESTMENT_EXPENSES: str = "Investment Expenses"
    MINING: str = "Mining"
    STAKING: str = "Staking"
    WAGES: str = "Wages"


_TEMPLATE_SHEETS_TO_KEEP: Set[str] = {f"__{item.value}" for item in SheetNames}

_SHEET_TO_TYPES: Dict[str, Tuple[TransactionType, ...]] = {
    SheetNames.AIRDROPS.value: (TransactionType.AIRDROP,),
    SheetNames.CAPITAL_GAINS.value: (TransactionType.SELL,),
    SheetNames.DONATIONS.value: (TransactionType.DONATE,),
    SheetNames.GIFTS.value: (TransactionType.GIFT,),
    SheetNames.HARDFORKS.value: (TransactionType.HARDFORK,),
    SheetNames.INCOME.value: (TransactionType.INCOME,),
    SheetNames.INTEREST.value: (TransactionType.INTEREST,),
    SheetNames.INVESTMENT_EXPENSES.value: (
        TransactionType.FEE,
        TransactionType.MOVE,
    ),
    SheetNames.MINING.value: (TransactionType.MINING,),
    SheetNames.STAKING.value: (TransactionType.STAKING,),
    SheetNames.WAGES.value: (TransactionType.WAGES,),
}

_TYPE_TO_SHEET: Dict[TransactionType, str] = {
    transaction_type: sheet_name for sheet_name, transaction_types in _SHEET_TO_TYPES.items() for transaction_type in transaction_types
}


class Generator(AbstractODSGenerator):
    MIN_ROWS: int = 20
    MAX_COLUMNS: int = 20
    OUTPUT_FILE: str = "tax_report_us.ods"

    HEADER_ROWS = 7

    def generate(
        self,
        country: AbstractCountry,
        years_2_accounting_method_names: Dict[int, str],
        asset_to_computed_data: Dict[str, ComputedData],
        output_dir_path: str,
        output_file_prefix: str,
        from_date: date,
        to_date: date,
        generation_language: str,
    ) -> None:
        row_indexes: Dict[str, int] = {sheet_name.value: self.HEADER_ROWS for sheet_name in SheetNames}

        if not isinstance(asset_to_computed_data, Dict):
            raise RP2TypeError(f"Parameter 'asset_to_computed_data' has non-Dict value {asset_to_computed_data}")

        template_path: str = self._get_template_path("tax_report_us", country, generation_language)

        output_file: Any
        output_file = self._initialize_output_file(
            country=country,
            legend_data=[],
            years_2_accounting_method_names=years_2_accounting_method_names,
            output_dir_path=output_dir_path,
            output_file_prefix=output_file_prefix,
            output_file_name=self.OUTPUT_FILE,
            template_path=template_path,
            template_sheets_to_keep=_TEMPLATE_SHEETS_TO_KEEP,
            from_date=from_date,
            to_date=to_date,
        )

        asset: str
        computed_data: ComputedData
        for asset, computed_data in asset_to_computed_data.items():
            if not isinstance(asset, str):
                raise RP2TypeError(f"Parameter 'asset' has non-string value {asset}")
            ComputedData.type_check("computed_data", computed_data)
            self.__generate(output_file, asset, computed_data.gain_loss_set, row_indexes)

        # Mark sheets that were not written to
        sheet_indexes_to_remove: List[int] = []
        index: int = 0
        sheet_name: str
        for sheet_name in output_file.sheets.names():
            if sheet_name != "Legend" and row_indexes[sheet_name] == Generator.HEADER_ROWS:
                sheet_indexes_to_remove.append(index)
            index += 1

        # Remove sheets that were marked for removal
        for index in reversed(sheet_indexes_to_remove):
            del output_file.sheets[index]

        output_file.save()
        LOGGER.info("Plugin '%s' output: %s", __name__, Path(output_file.docname).resolve())

    def __generate(self, output_file: Any, asset: str, gain_loss_set: GainLossSet, row_indexes: Dict[str, int]) -> None:
        sheet: Any
        for sheet in output_file.sheets:
            if sheet.name == "Legend":
                continue
            sheet_types: Tuple[TransactionType, ...] = _SHEET_TO_TYPES[sheet.name]
            for sheet_type in sheet_types:
                sheet.append_rows(self.MIN_ROWS + gain_loss_set.get_transaction_type_count(sheet_type) + 1)

        border_suffix: str = "_border"
        for entry in gain_loss_set:
            gain_loss: GainLoss = cast(GainLoss, entry)
            sheet_type = gain_loss.taxable_event.transaction_type
            sheet = output_file.sheets[_TYPE_TO_SHEET[sheet_type]]
            row_index: int = row_indexes[sheet.name]
            current_taxable_event_fraction: int = gain_loss_set.get_taxable_event_fraction(gain_loss) + 1
            total_taxable_event_fractions: int = gain_loss_set.get_taxable_event_number_of_fractions(gain_loss.taxable_event)
            transaction_type: str = (
                f"{self._get_table_type_from_transaction(gain_loss.taxable_event)} / " f"{gain_loss.taxable_event.transaction_type.value.upper()}"
            )
            taxable_event_note: str = (
                f"{current_taxable_event_fraction}/"
                f"{total_taxable_event_fractions}: "
                f"{gain_loss.crypto_amount:.8f} of "
                f"{gain_loss.taxable_event.crypto_balance_change:.8f} "
                f"{asset}"
            )
            transparent_vs: str = f"transparent{border_suffix}"
            taxable_event_note_vs: str = f"taxable_event_note{border_suffix}"
            acquired_lot_note_vs: str = f"acquired_lot_note{border_suffix}"

            self._fill_cell(sheet, row_index, 0, gain_loss.crypto_amount, visual_style=transparent_vs, data_style="crypto")
            self._fill_cell(sheet, row_index, 1, gain_loss.asset, visual_style=transparent_vs)
            self._fill_cell(sheet, row_index, 3, gain_loss.taxable_event.timestamp.strftime("%m/%d/%Y"), visual_style=taxable_event_note_vs)
            self._fill_cell(sheet, row_index, 4, gain_loss.taxable_event_fiat_amount_with_fee_fraction, visual_style=taxable_event_note_vs, data_style="fiat")
            self._fill_cell(sheet, row_index, 6, "", visual_style=transparent_vs)
            self._fill_cell(sheet, row_index, 7, "", visual_style=transparent_vs)
            self._fill_cell(sheet, row_index, 8, gain_loss.fiat_gain, visual_style=transparent_vs, data_style="fiat")
            self._fill_cell(sheet, row_index, 9, transaction_type, visual_style=taxable_event_note_vs)
            self._fill_cell(sheet, row_index, 12, taxable_event_note, visual_style=taxable_event_note_vs)
            self._fill_cell(sheet, row_index, 13, gain_loss.taxable_event.unique_id, visual_style=taxable_event_note_vs)
            self._fill_cell(sheet, row_index, 14, "LONG" if gain_loss.is_long_term_capital_gains() else "SHORT", visual_style=taxable_event_note_vs)
            self._fill_cell(sheet, row_index, 15, gain_loss.taxable_event.timestamp, visual_style=taxable_event_note_vs)

            if gain_loss.acquired_lot:
                current_acquired_lot_fraction: int = gain_loss_set.get_acquired_lot_fraction(gain_loss) + 1
                total_acquired_lot_fractions: int = gain_loss_set.get_acquired_lot_number_of_fractions(gain_loss.acquired_lot)
                acquired_lot_note: str = (
                    f"{current_acquired_lot_fraction}/"
                    f"{total_acquired_lot_fractions}: "
                    f"{gain_loss.crypto_amount:.8f} of "
                    f"{gain_loss.acquired_lot.crypto_balance_change:.8f} "
                    f"{asset}"
                )
                self._fill_cell(sheet, row_index, 2, gain_loss.acquired_lot.timestamp.strftime("%m/%d/%Y"), visual_style=acquired_lot_note_vs)
                self._fill_cell(sheet, row_index, 5, gain_loss.fiat_cost_basis, visual_style=acquired_lot_note_vs, data_style="fiat")
                self._fill_cell(sheet, row_index, 10, acquired_lot_note, visual_style=acquired_lot_note_vs)
                self._fill_cell(sheet, row_index, 11, gain_loss.acquired_lot.unique_id, visual_style=acquired_lot_note_vs)
            else:
                self._fill_cell(sheet, row_index, 2, "", visual_style=transparent_vs)
                self._fill_cell(sheet, row_index, 5, "", visual_style=transparent_vs)
                self._fill_cell(sheet, row_index, 10, "", visual_style=transparent_vs)
                self._fill_cell(sheet, row_index, 11, "", visual_style=transparent_vs)

            border_suffix = ""
            row_indexes[sheet.name] = row_index + 1



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2TypeError


class TestRP2Decimal(unittest.TestCase):
    def setUp(self) -> None:
        self.maxDiff = None  # pylint: disable=invalid-name

    def test_rp2_decimal(self) -> None:
        decimal1: RP2Decimal = RP2Decimal("0.2")
        decimal2: RP2Decimal = RP2Decimal("0.2")
        decimal3: RP2Decimal = RP2Decimal("0.20000000000001")
        decimal4: RP2Decimal = RP2Decimal("0.2000000000001")

        decimal5: RP2Decimal = RP2Decimal("0.1")
        decimal6: RP2Decimal = RP2Decimal("0.1")
        decimal7: RP2Decimal = RP2Decimal("0.10000000000001")
        decimal8: RP2Decimal = RP2Decimal("0.1000000000001")

        zero_point_twelve = RP2Decimal("0.12")
        one: RP2Decimal = RP2Decimal("1")
        two: RP2Decimal = RP2Decimal("2")
        three: RP2Decimal = RP2Decimal("3")
        nine: RP2Decimal = RP2Decimal("9")

        self.assertTrue(decimal1 == decimal2)
        self.assertTrue(decimal1 == decimal3)
        self.assertFalse(decimal1 == decimal4)

        self.assertFalse(decimal1 != decimal2)
        self.assertFalse(decimal1 != decimal3)
        self.assertTrue(decimal1 != decimal4)

        self.assertFalse(decimal1 > decimal2)
        self.assertFalse(decimal1 > decimal3)
        self.assertFalse(decimal1 > decimal4)

        self.assertTrue(decimal1 >= decimal2)
        self.assertTrue(decimal1 >= decimal3)
        self.assertFalse(decimal1 >= decimal4)

        self.assertFalse(decimal1 < decimal2)
        self.assertFalse(decimal1 < decimal3)
        self.assertTrue(decimal1 < decimal4)

        self.assertTrue(decimal1 <= decimal2)
        self.assertTrue(decimal1 <= decimal3)
        self.assertTrue(decimal1 <= decimal4)

        # Check that overridden operators return RP2Decimal (these tests would fail without the custom precision built into RP2Decimal comparison operators)
        self.assertTrue(decimal5 + decimal6 == decimal1)
        self.assertTrue(decimal6 + decimal5 == decimal1)
        self.assertTrue(decimal5 + decimal7 == decimal1)
        self.assertTrue(decimal7 + decimal5 == decimal1)
        self.assertFalse(decimal5 + decimal8 == decimal1)
        self.assertFalse(decimal8 + decimal5 == decimal1)
        self.assertTrue(decimal5 + decimal7 == decimal5 + decimal6)
        self.assertFalse(decimal5 + decimal8 == decimal5 + decimal6)

        self.assertTrue(decimal1 - decimal2 == ZERO)
        self.assertTrue(decimal3 - decimal1 == ZERO)
        self.assertTrue(decimal4 - decimal1 > ZERO)

        self.assertTrue(decimal1 * three * decimal2 == zero_point_twelve)
        self.assertTrue(decimal1 * three * decimal3 == zero_point_twelve)
        self.assertTrue(decimal1 * three * decimal4 > zero_point_twelve)

        self.assertTrue(decimal2 / decimal1 == one)
        self.assertTrue(decimal3 / decimal1 == one)
        self.assertTrue(decimal4 / decimal1 > one)

        self.assertTrue(three // two == one)
        self.assertTrue((three // two) + decimal3 - decimal1 == one)
        self.assertTrue((three // two) + decimal4 - decimal1 > one)

        self.assertTrue(pow(three, two) == nine)
        self.assertTrue(pow(three, two) + decimal3 - decimal1 == nine)
        self.assertTrue(pow(three, two) + decimal4 - decimal1 > nine)

        self.assertTrue(nine % two == one)
        self.assertTrue(nine % two + decimal3 - decimal1 == one)
        self.assertTrue(nine % two + decimal4 - decimal1 > one)

    def test_bad_rp2_decimal(self) -> None:
        # pylint: disable=pointless-statement
        one: RP2Decimal = RP2Decimal("1")

        # Test comparison operators
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            one == 1
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            1 == one
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            one != -1.1
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            1.1 != one
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            one > 1
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            1 > one
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            one < 1
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            1 < one
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            one >= 1
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            1 >= one
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            one <= 1
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            1 <= one

        # Test arithmetic operators
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            one + 1
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            1 + one
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            one - 1
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            1 - one
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            one * 1
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            1 * one
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            one / 1
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            1 / one
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            one // 1
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            1 // one

        # Test power
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            one**1
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            pow(one, 1)
        with self.assertRaisesRegex(RP2TypeError, "Modulo has non-Decimal value "):
            pow(one, one, 1)

        # Test modulo
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            one % 1
        with self.assertRaisesRegex(RP2TypeError, "Operand has non-Decimal value "):
            1 % one


if __name__ == "__main__":
    unittest.main()



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import sys
from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from configparser import ConfigParser
from pathlib import Path
from typing import Any

from jsonschema import validate

from rp2.configuration import Keyword
from rp2.configuration_schema import CONFIGURATION_SCHEMA
from rp2.logger import LOGGER

_VERSION: str = "0.1.0"


def rp2_configuration_translator() -> None:
    args: Namespace
    parser: ArgumentParser

    parser = _setup_argument_parser()
    args = parser.parse_args()

    json_file_name: str = args.json_file
    ini_file_name: str = args.output_file if args.output_file else f"{str(Path(json_file_name).parent/Path(json_file_name).stem)}.ini"

    if not Path(json_file_name).exists():
        print(f"Configuration file '{json_file_name}' not found")
        parser.print_help()
        sys.exit(1)

    if Path(ini_file_name).exists() and not args.force_overwrite:
        print(f"Output file '{ini_file_name}' exists")
        parser.print_help()
        sys.exit(1)

    try:
        json_configuration: Any
        with open(json_file_name, encoding="utf-8") as json_file:
            # This json_configuration is validated by jsonschema, so we can disable static type checking for it:
            # it adds complexity but not much value over jsonschema checks
            json_configuration = json.load(json_file)
            validate(instance=json_configuration, schema=CONFIGURATION_SCHEMA)

        ini_object = ConfigParser()

        ini_object[Keyword.GENERAL.value] = {
            Keyword.ASSETS.value: ", ".join(json_configuration[Keyword.ASSETS.value]),
            Keyword.EXCHANGES.value: ", ".join(json_configuration[Keyword.EXCHANGES.value]),
            Keyword.HOLDERS.value: ", ".join(json_configuration[Keyword.HOLDERS.value]),
        }
        if Keyword.GENERATORS.value in json_configuration:
            ini_object[Keyword.GENERAL.value][Keyword.GENERATORS.value] = json_configuration[Keyword.GENERATORS.value]
        if Keyword.ACCOUNTING_METHODS.value in json_configuration:
            ini_object[Keyword.ACCOUNTING_METHODS.value] = json_configuration[Keyword.ACCOUNTING_METHODS.value]
        ini_object[Keyword.IN_HEADER.value] = json_configuration[Keyword.IN_HEADER.value]
        ini_object[Keyword.OUT_HEADER.value] = json_configuration[Keyword.OUT_HEADER.value]
        ini_object[Keyword.INTRA_HEADER.value] = json_configuration[Keyword.INTRA_HEADER.value]

        with open(ini_file_name, "w", encoding="utf-8") as ini_file:
            ini_object.write(ini_file)

    except Exception:  # pylint: disable=broad-except
        LOGGER.exception("Fatal exception occurred:")
        sys.exit(1)

    LOGGER.info("Input JSON configuration file: %s", json_file_name)
    LOGGER.info("Output INI configuration file: %s", ini_file_name)
    LOGGER.info("Done")


def _setup_argument_parser() -> ArgumentParser:
    parser: ArgumentParser = ArgumentParser(
        description=("Convert a JSON-format RP2 configuration file to INI format"),
        formatter_class=RawTextHelpFormatter,
    )

    parser.add_argument(
        "-f",
        "--force-overwrite",
        action="store_true",
        help="Write the output file even if it already exists",
    )
    parser.add_argument(
        "-o",
        "--output_file",
        action="store",
        default="",
        help="Write INI format configuration to OUTPUT_FILE",
        metavar="OUTPUT_FILE",
        type=str,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"RP2 configuration translator {_VERSION} (https://github.com/eprbell/rp2)",
        help="Print version",
    )
    parser.add_argument(
        "json_file",
        action="store",
        help="JSON configuration file",
        metavar="JSON_FILE",
        type=str,
    )

    return parser



# Copyright 2021 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import date
from typing import cast

from rp2.configuration import MAX_DATE, MIN_DATE, Configuration
from rp2.entry_types import EntrySetType
from rp2.rp2_error import RP2TypeError
from rp2.transaction_set import TransactionSet


class InputData:
    @classmethod
    def type_check(cls, name: str, instance: "InputData") -> "InputData":
        Configuration.type_check_parameter_name(name)
        if not isinstance(instance, cls):
            raise RP2TypeError(f"Parameter '{name}' is not of type {cls.__name__}: {instance}")
        return instance

    def __init__(
        self,
        asset: str,
        unfiltered_in_transaction_set: TransactionSet,
        unfiltered_out_transaction_set: TransactionSet,
        unfiltered_intra_transaction_set: TransactionSet,
        from_date: date = MIN_DATE,
        to_date: date = MAX_DATE,
    ):
        self.__asset: str = Configuration.type_check_string("asset", asset)
        self.__unfiltered_in_transaction_set: TransactionSet = TransactionSet.type_check(
            "in_transaction_set", unfiltered_in_transaction_set, EntrySetType.IN, self.asset, False
        )
        self.__unfiltered_out_transaction_set: TransactionSet = TransactionSet.type_check(
            "out_transaction_set", unfiltered_out_transaction_set, EntrySetType.OUT, self.asset, True
        )
        self.__unfiltered_intra_transaction_set: TransactionSet = TransactionSet.type_check(
            "intra_transaction_set", unfiltered_intra_transaction_set, EntrySetType.INTRA, self.asset, True
        )
        if not isinstance(from_date, date):
            raise RP2TypeError("Parameter 'from_date' is not of type date")
        if not isinstance(to_date, date):
            raise RP2TypeError("Parameter 'to_date' is not of type date")

        self.__filtered_in_transaction_set: TransactionSet = cast(
            TransactionSet, self.__unfiltered_in_transaction_set.duplicate(from_date=from_date, to_date=to_date)
        )
        self.__filtered_out_transaction_set: TransactionSet = cast(
            TransactionSet, self.__unfiltered_out_transaction_set.duplicate(from_date=from_date, to_date=to_date)
        )
        self.__filtered_intra_transaction_set: TransactionSet = cast(
            TransactionSet, self.__unfiltered_intra_transaction_set.duplicate(from_date=from_date, to_date=to_date)
        )

    @property
    def asset(self) -> str:
        return self.__asset

    @property
    def unfiltered_in_transaction_set(self) -> TransactionSet:
        return self.__unfiltered_in_transaction_set

    @property
    def unfiltered_out_transaction_set(self) -> TransactionSet:
        return self.__unfiltered_out_transaction_set

    @property
    def unfiltered_intra_transaction_set(self) -> TransactionSet:
        return self.__unfiltered_intra_transaction_set

    @property
    def filtered_in_transaction_set(self) -> TransactionSet:
        return self.__filtered_in_transaction_set

    @property
    def filtered_out_transaction_set(self) -> TransactionSet:
        return self.__filtered_out_transaction_set

    @property
    def filtered_intra_transaction_set(self) -> TransactionSet:
        return self.__filtered_intra_transaction_set
